import factory
import rstr

from wbfdm.models import (
    Classification,
    ClassificationGroup,
    InstrumentClassificationThroughModel,
)


class ClassificationGroupFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ClassificationGroup
        django_get_or_create = ("is_primary",)

    name = factory.Sequence(lambda n: f"Classification Group {n}")
    is_primary = True
    max_depth = 1
    code_level_digits = 2


class ClassificationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Classification

    code_aggregated = factory.LazyAttribute(lambda o: rstr.xeger("([0-9]{8})"))
    name = factory.Sequence(lambda n: f"Classification {n}")
    # parent = factory.SubFactory("wbfdm.factories.ParentClassificationFactory", parent=None)
    group = factory.SubFactory(ClassificationGroupFactory)
    level = 0
    # level_representation = factory.LazyAttribute(lambda x: f"Level {x.level}")


class ParentClassificationFactory(ClassificationFactory):
    parent = None


class InstrumentClassificationThroughModelFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = InstrumentClassificationThroughModel
        django_get_or_create = ("instrument", "classification")
        skip_postgeneration_save = True

    instrument = factory.SubFactory("wbfdm.factories.InstrumentFactory")
    classification = factory.SubFactory("wbfdm.factories.ClassificationFactory")
    is_favorite = factory.Faker("pybool")
    reason = factory.Faker("paragraph")
    pure_player = factory.Faker("pybool")
    top_player = factory.Faker("pybool")

    @factory.post_generation
    def related_instruments(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for instrument in extracted:
                self.related_instruments.add(instrument)



from typing import TYPE_CHECKING, Generator, Iterable, TypeVar

import pandas as pd
from plotly import graph_objects as go

from ..utils import normalize_series

if TYPE_CHECKING:
    from wbfdm.analysis.technical_analysis.technical_analysis import TechnicalAnalysis


T = TypeVar("T", bound=go.Candlestick | go.Ohlc)


class TechnicalAnalysisTraceFactory:
    def __init__(self, ta: "TechnicalAnalysis"):
        self.ta = ta

    def _line_trace(
        self,
        key: str,
        prefix: str,
        base_series: pd.Series | None = None,
        percent_format: bool = False,
        line_options: dict | None = None,
        show_estimates: bool = True,
    ) -> Generator[go.Scatter, None, None]:
        # Parametrization
        series = self.ta.df[key]

        # get the calculated time series. By default all value are considered non estimated
        try:
            calculated = self.ta.df["calculated"]
        except (KeyError, UserWarning):
            calculated = pd.Series(False, index=self.ta.df.index)
        name = str(self.ta.instrument)
        line_options = line_options if line_options else {}
        if not (series := normalize_series(series, base_series=base_series)).empty:
            real_series = series.loc[~calculated]
            calculated_series = series.loc[calculated]
            inception_date = None
            if not real_series.empty:
                inception_date = real_series.index.min()
                # Prepare chart options
                text = []
                if base_series is not None and not base_series.empty:
                    prefix = f"Normalized {prefix}"
                format_template = "{y:.2%}" if percent_format else "{y:.2f}"
                hovertemplate = "<b>" + prefix + "</b>: %" + format_template
                # If a base series is present, we add a line into the hover template to show the difference
                if base_series is not None and not base_series.empty:
                    hovertemplate += "<br>∆ : %{text}"
                    text = [format_template.format(y=y) for y in (real_series - base_series).dropna().values]

                yield go.Scatter(
                    x=real_series.index,
                    y=real_series.tolist(),
                    line=line_options,
                    name=name,
                    legendgroup=name,
                    hovertemplate=hovertemplate,
                    text=text,
                )
            if show_estimates and not calculated_series.empty:
                dash_line_options = {**line_options, "dash": "dashdot"}
                if not inception_date:
                    inception_date = calculated_series.index.min()
                backtesting_series = calculated_series.loc[calculated_series.index < inception_date]
                other_calculated_series = calculated_series.loc[calculated_series.index >= inception_date]
                if not backtesting_series.empty:
                    yield go.Scatter(
                        x=backtesting_series.index,
                        y=backtesting_series.tolist(),
                        line=dash_line_options,
                        name="Backtesting " + name,
                        legendgroup=name,
                    )
                if not other_calculated_series.empty:
                    yield go.Scatter(
                        x=other_calculated_series.index,
                        y=other_calculated_series.tolist(),
                        line=line_options if real_series.shape[0] > 1 else dash_line_options,
                        name="Estimated " + name,
                        mode="markers" if real_series.shape[0] > 1 else "lines",
                        legendgroup=name,
                    )

    def _ohlc_trace(self, graph_object: type[T]) -> Iterable[T]:
        df = self.ta.df
        name = str(self.ta.instrument)
        yield graph_object(
            x=df["close"].index,
            close=df["close"].values,
            open=df["open"].values,
            high=df["high"].values,
            low=df["low"].values,
            name=name,
            legendgroup=name,
        )

    def performance_summary_trace(self, bar_options: dict | None = None, **kwargs) -> go.Bar:
        summaries = {
            "36 Months": self.ta.get_performance_months(months=36),
            "24 Months": self.ta.get_performance_months(months=24),
            "12 Months": self.ta.get_performance_months(months=12),
            "YTD": self.ta.get_performance_year_to_date(),
            "6 Months": self.ta.get_performance_months(months=6),
            "3 Months": self.ta.get_performance_months(months=3),
            "1 Months": self.ta.get_performance_months(months=1),
        }

        yield go.Bar(
            x=list(reversed(summaries.keys())),
            y=list(reversed(summaries.values())),
            text=list(map(lambda x: f"{x:.2%}", reversed(summaries.values()))),
            textposition="auto",
            marker=bar_options,
        )

    def volume_trace(self) -> go.Bar:
        df = self.ta.df
        df["color_volume"] = "green"
        df.loc[df.volume < 0, "color_volume"] = "red"
        bar_chart_name = "Inflow/Outflow" if self.ta.instrument.is_managed else "Volume"
        return go.Bar(
            x=df.index,
            y=df["volume"],
            name=bar_chart_name,
            hovertemplate="<b>" + bar_chart_name + "</b><br>%{x}<br>%{y:.4s}<extra></extra>",
            marker={"color": df.color_volume, "opacity": 0.4},
            yaxis="y2",
        )

    def close_trace(
        self,
        base_series: pd.Series | None = None,
        line_options: dict | None = None,
        show_estimates: bool = True,
        **kwargs,
    ) -> Generator[go.Scatter, None, None]:
        yield from self._line_trace(
            "close", "Close", base_series=base_series, line_options=line_options, show_estimates=show_estimates
        )

    def sma_trace(
        self, window: int, base_series: pd.Series | None = None, line_options: dict | None = None, **kwargs
    ) -> Generator[go.Scatter, None, None]:
        self.ta.add_sma(window)
        yield from self._line_trace(
            f"close_{window}_sma", f"SMA {window}", base_series=base_series, line_options=line_options
        )

    def return_trace(
        self, base_series: pd.Series | None = None, line_options: dict | None = None, **kwargs
    ) -> Generator[go.Scatter, None, None]:
        self.ta.add_cumulative_return()
        yield from self._line_trace(
            "cum-ret", "Return", percent_format=True, base_series=base_series, line_options=line_options
        )

    def log_return_trace(
        self, base_series: pd.Series | None = None, line_options: dict | None = None, **kwargs
    ) -> Generator[go.Scatter, None, None]:
        self.ta.add_cumulative_return("log")
        yield from self._line_trace(
            "cum-log-ret", "Log-Return", percent_format=True, base_series=base_series, line_options=line_options
        )

    def drawdown_trace(
        self, base_series: pd.Series | None = None, line_options: dict | None = None, **kwargs
    ) -> Generator[go.Scatter, None, None]:
        self.ta.add_drawdown()
        yield from self._line_trace("drawdown", "Drawdown", base_series=base_series, line_options=line_options)

    def candlestick_trace(self, **kwargs) -> go.Candlestick:
        yield from self._ohlc_trace(go.Candlestick)

    def ohlc_trace(self, **kwargs) -> go.Ohlc:
        yield from self._ohlc_trace(go.Ohlc)



from contextlib import suppress
from datetime import date
from itertools import batched
from typing import Iterator

import pypika as pk
from django.db import ProgrammingError, connections
from pypika import functions as fn
from pypika.enums import Order, SqlTypes
from pypika.terms import LiteralValue
from wbcore.contrib.dataloader.dataloaders import Dataloader
from wbcore.contrib.dataloader.utils import dictfetchall

from wbfdm.contrib.qa.dataloaders.utils import SOURCE_DS2
from wbfdm.dataloaders.protocols import AdjustmentsProtocol
from wbfdm.dataloaders.types import AdjustmentDataDict


class DatastreamAdjustmentsDataloader(AdjustmentsProtocol, Dataloader):
    def adjustments(self, from_date: date | None = None, to_date: date | None = None) -> Iterator[AdjustmentDataDict]:
        lookup = {k: v for k, v in self.entities.values_list("dl_parameters__adjustments__parameters", "id")}

        adj = pk.Table("DS2Adj")
        adj_date = fn.Cast(adj.AdjDate, SqlTypes.DATE).as_("adjustment_date")
        adj_end_date = fn.Cast(adj.EndAdjDate, SqlTypes.DATE).as_("adjustment_end_date")

        infocode = pk.Table("#ds2infocode")

        query = (
            pk.MSSQLQuery.select(
                adj.InfoCode.as_("external_identifier"),
                fn.Concat(adj.InfoCode, "_", adj_date).as_("id"),
                adj_date,
                adj_end_date,
                SOURCE_DS2,
                adj.AdjFactor.as_("adjustment_factor"),
                adj.CumAdjFactor.as_("cumulative_adjustment_factor"),
            )
            .from_(adj)
            .where(adj.AdjType == 2)
            .where(adj.InfoCode.isin([LiteralValue("select infocode from #ds2infocode")]))
            .orderby(adj.AdjDate, order=Order.desc)
        )

        if from_date:
            query = query.where(adj.AdjDate >= from_date)

        if to_date:
            query = query.where(adj.AdjDate <= to_date)

        with connections["qa"].cursor() as cursor:
            # we suppress an error here, because if the temporary table already exists
            # then we do not want to fail. It should not fail, but if a previous run did
            # not clean up the table properly, then at least we do not get stuck here
            with suppress(ProgrammingError):
                cursor.execute(
                    pk.MSSQLQuery.create_table(infocode).columns(pk.Column("infocode", SqlTypes.INTEGER)).get_sql()
                )
                for batch in batched(lookup.keys(), 1000):
                    placeholders = ",".join(map(lambda x: f"({x})", batch))
                    cursor.execute("insert into #ds2infocode values %s;", (placeholders,))

            cursor.execute(query.get_sql())

            for row in dictfetchall(cursor, AdjustmentDataDict):
                row["instrument_id"] = lookup[row["external_identifier"]]
                yield row

            # here we remove the temporary table again to avoid data spillage
            # cursor.execute(pk.MSSQLQuery.drop_table(infocode).get_sql())



import pandas as pd
from django.contrib.messages import warning
from django.utils.functional import cached_property
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf
from wbcore.contrib.pandas.utils import (
    override_number_to_decimal,
    override_number_to_integer_without_decorations,
    override_number_to_percent,
    override_number_to_x,
    pct_change_with_negative_values,
    sanitize_fields,
)
from wbcore.metadata.configs.endpoints import NoEndpointViewConfig

from wbfdm.enums import Financial
from wbfdm.models.instruments import Instrument
from wbfdm.viewsets.configs.display import (
    FinancialSummaryDisplayViewConfig,
)

from ..mixins import InstrumentMixin


class FinancialSummary(InstrumentMixin, ExportPandasAPIViewSet):
    queryset = Instrument.objects.none()
    display_config_class = FinancialSummaryDisplayViewConfig
    ordering_fields = "__all__"
    endpoint_config_class = NoEndpointViewConfig

    def get_queryset(self):
        return Instrument.objects.filter(id=self.instrument.id)

    def get_pandas_fields(self, request):
        return pf.PandasFields(
            fields=[
                pf.PKField(key="id", label="ID"),
                pf.CharField(key="label", label="Financial"),
                pf.JsonField(key="_overwrites", label="Overwrites"),
                *[pf.FloatField(key=k, label=k, precision=1) for k in self.fiscal_columns],
            ]
        )

    def add_messages(self, request, instance=None, **kwargs):
        super().add_messages(request, instance=instance, **kwargs)
        if self.df.empty:
            warning(request, "There is no data available for the Financial Summary table", extra_tags="auto_close=0")

    def get_dataframe(self, request, queryset, **kwargs):
        # Get all necessary data from the dataloader and load a dataframe
        df = pd.DataFrame(
            queryset.dl.financials(
                values=self.FINANCIAL_VALUES,
                from_index=-5,
                to_index=3,
            )
        )
        if not df.empty:
            # Pivot the data
            df = df.pivot_table(
                columns="financial",
                index=["year", "period_end_date", "estimate"],
                values="value",
            ).rename_axis(columns=None)

            sanitize_fields(df, map(lambda enum: enum.value, self.FINANCIAL_VALUES))

            # Compute all necessary fields
            df["revenue_growth"] = df["revenue"].pct_change() * 100
            df["gross_profit_pct"] = df["gross_profit"] / df["revenue"] * 100
            df["ebitda_pct"] = df["ebitda"] / df["revenue"] * 100
            df["ebit_pct"] = df["ebit"] / df["revenue"] * 100
            df["net_income_pct"] = df["net_income"] / df["revenue"] * 100
            df["eps_growth"] = pct_change_with_negative_values(df, "eps") * 100

            df["net_debt_ebitda"] = df["net_debt"] / df["ebitda"]
            df["debt_assets"] = df["total_debt"] / df["total_assets"] * 100
            df["debt_equity"] = df["total_debt"] / df["total_assets"] * 100

            df["interest_coverage_ratio"] = df["ebit"] / df["interest_expense"]
            df["free_cash_flow_per_share"] = df["free_cash_flow"] / df["shares_outstanding"]
            df["free_cash_flow_per_share_growth"] = (
                pct_change_with_negative_values(df, "free_cash_flow_per_share") * 100
            )

            # Normalize data
            df["revenue"] = df["revenue"] / 1_000_000
            df["net_income"] = df["net_income"] / 1_000_000

            # Sort the columns into the desired order
            # Reset the index to get the period end date as a column
            # Pivot back to have the dates on top
            df = df[self.FIELDS]

            # Reset the 2 indices and transpose back
            df = df.reset_index(level=[0, 2]).sort_index()

            # Adjust the columns to be in a different format
            df.index = df.index.map(lambda x: x.strftime("%b/%y"))
            max_row = 8
            if df.shape[0] > max_row:
                df = df.iloc[1:]  # remove first row
                df = df.iloc[0 : min([df.shape[0], max_row])]  # keep only 8 row maximum

            self._estimate_columns = df["estimate"].to_dict()
            df = df.drop(columns=["estimate"], errors="ignore")
        return df

    def manipulate_dataframe(self, df):
        if not df.empty:
            df = df.T
            # Add labels for human readable output
            df["label"] = self.LABELS

            override_number_to_percent(
                df,
                *list(
                    map(
                        lambda x: df.index == x,
                        [
                            "revenue_growth",
                            "gross_profit_pct",
                            "ebitda_pct",
                            "ebit_pct",
                            "net_income_pct",
                            "eps_growth",
                            "roe",
                            "roic",
                            "roc",
                            "roa",
                            "debt_equity",
                            "debt_assets",
                            "free_cash_flow_per_share_growth",
                        ],
                    )
                ),
            )

            override_number_to_x(
                df,
                *list(
                    map(
                        lambda x: df.index == x,
                        [
                            "net_debt_ebitda",
                            "interest_coverage_ratio",
                        ],
                    )
                ),
            )

            override_number_to_decimal(
                df,
                *list(
                    map(
                        lambda x: df.index == x,
                        [
                            "revenue",
                            "net_income",
                        ],
                    )
                ),
            )

            override_number_to_integer_without_decorations(
                df,
                *list(
                    map(
                        lambda x: df.index == x,
                        [
                            "year",
                        ],
                    )
                ),
            )

            df = df.reset_index(names="id")
        return df

    @cached_property
    def fiscal_columns(self) -> list:
        """Returns the fiscal columns from the dataframe"""
        return self.df.columns.difference(["label", "_overwrites", "id"]).to_list()

    @cached_property
    def estimate_columns(self) -> dict:
        """Returns a dictionary with the estimate column for each fiscal column
        The _estimate_columns will be set if the dataframe is constructed.
        """
        return getattr(self, "_estimate_columns", {})

    @cached_property
    def FINANCIAL_VALUES(self) -> list[Financial]:  # noqa
        return [
            Financial.REVENUE,  # SAL
            Financial.GROSS_PROFIT,  # GRI
            Financial.EBITDA,  # EBT
            Financial.EBIT,  # EBI
            Financial.NET_INCOME,  # NET
            Financial.EPS,  # EPS
            Financial.SHAREHOLDERS_EQUITY,  # SHE
            Financial.TOTAL_ASSETS,  # TAS
            Financial.TAX_RATE,  # TAX
            Financial.RETURN_ON_INVESTED_CAPITAL,  # RIC
            Financial.NET_DEBT,  # NDT
            Financial.TOTAL_DEBT,  # TDT
            Financial.INTEREST_EXPENSE,  # INE
            Financial.FREE_CASH_FLOW,  # FCF
            Financial.SHARES_OUTSTANDING,
            Financial.CURRENT_LIABILITIES,  # CRL
            Financial.CASH_EQUIVALENTS,
            Financial.RETURN_ON_EQUITY,
            Financial.RETURN_ON_ASSETS,
            Financial.RETURN_ON_CAPITAL,
            Financial.RETURN_ON_INVESTED_CAPITAL,
        ]

    @cached_property
    def FIELDS(self) -> list[str]:  # noqa
        return [
            "revenue",
            "revenue_growth",
            "gross_profit_pct",
            "ebitda_pct",
            "ebit_pct",
            "net_income_pct",
            "net_income",
            "eps",
            "eps_growth",
            "roe",
            "roa",
            "roc",
            "roic",
            "net_debt_ebitda",
            "debt_assets",
            "debt_equity",
            "interest_coverage_ratio",
            "free_cash_flow_per_share",
            "free_cash_flow_per_share_growth",
        ]

    @property
    def LABELS(self) -> list[str]:  # noqa
        currency_key = self.instrument.currency.key if self.instrument.currency else "N.A."
        return [
            f"in {currency_key} MN",
            "Revenue",
            "Y/Y Change",
            "Gross Margin",
            "EBITDA Margin",
            "EBIT Margin",
            "Net Profit Margin",
            "Net Profit",
            "EPS",
            "Y/Y Change",
            "ROE",
            "ROA",
            "ROC",
            "ROIC",
            "Net Debt/EBITDA",
            "D/A",
            "D/E",
            "Int. Cov. Ratio",
            "FCF per share",
            "Y/Y Change",
        ]



from typing import Optional

from django.utils.translation import gettext_lazy as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class ClassificationDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label="Title"),
                dp.Field(key="code_aggregated", label="Code"),
                dp.Field(key="parent", label="Parent"),
                dp.Field(key="group", label="Group"),
                dp.Field(key="height", label="Height"),
                dp.Field(key="level", label="Level"),
                dp.Field(key="level_representation", label="Level Representation"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["name", "code_aggregated"],
                ["parent", "group"],
                ["height", "level_representation"],
                [repeat_field(2, "description")],
                [repeat_field(2, "childs_section")],
                [repeat_field(2, "instruments_section")],
            ],
            [
                create_simple_section(
                    "childs_section", _("Child Classifications"), [["childs"]], "childs", collapsed=False
                ),
                create_simple_section(
                    "instruments_section", _("Instruments"), [["instruments"]], "instruments", collapsed=True
                ),
            ],
        )


class ClassificationGroupDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label="Title"),
                dp.Field(key="is_primary", label="Primary"),
                dp.Field(key="max_depth", label="Max Depth"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(2, "name")],
                ["is_primary", "max_depth"],
                [repeat_field(2, "classifications_section")],
            ],
            [
                create_simple_section(
                    "classifications_section",
                    _("Classification"),
                    [["classifications"]],
                    "classifications",
                    collapsed=False,
                )
            ],
        )


class InstrumentClassificationThroughDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="instrument", label="Instrument") if "instrument_id" not in self.view.kwargs else None,
                dp.Field(key="classification", label="Classification")
                if "classification_id" not in self.view.kwargs
                else None,
                dp.Field(key="is_favorite", label="Favorite"),
                dp.Field(key="pure_player", label="Pure Player"),
                dp.Field(key="top_player", label="Top Player"),
                dp.Field(key="reason", label="Reason"),
                dp.Field(key="percent_of_revenue", label="% of revenue"),
                dp.Field(key="tags", label="Tags"),
                dp.Field(key="related_instruments", label="RelatedInstruments"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [
                    "instrument" if "instrument_id" not in self.view.kwargs else ".",
                    "classification" if "classification_id" not in self.view.kwargs else ".",
                ],
                ["is_favorite", "top_player"],
                ["pure_player", "percent_of_revenue"],
                [repeat_field(2, "reason")],
                [repeat_field(2, "tags")],
                [repeat_field(2, "related_instruments_section")],
            ],
            [
                create_simple_section(
                    "related_instruments_section",
                    _("Related Instruments"),
                    [["related_instruments"]],
                    "related_instruments",
                    collapsed=False,
                )
            ],
        )


class ClassificationInstrumentRelatedInstrumentDisplayConfig(DisplayViewConfig):
    def get_list_display(self):
        return dp.ListDisplay(
            fields=[
                dp.Field(key="related_instrument", label="Related Instrument"),
                dp.Field(key="related_instrument_type", label="Type"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [
                    "related_instrument",
                    "related_instrument_type",
                ]
            ]
        )



from datetime import date

from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import DatePreference, IntPreference, StringPreference

fdm = Section("wbfdm")


@global_preferences_registry.register
class DefaultClassificationGroup(IntPreference):
    section = fdm
    name = "default_classification_group"
    default = 0

    verbose_name = "Default Classification Group"


@global_preferences_registry.register
class DefaultStartDateHistoricalImport(DatePreference):
    section = fdm
    name = "default_start_date_historical_import"
    default = date(2015, 1, 1)

    verbose_name = "Default Start Date"
    help_text = "Default start date in historical import"


@global_preferences_registry.register
class NonTickerWords(StringPreference):
    section = fdm
    name = "non_ticker_words"
    default = ""

    verbose_name = "Non Ticker Words"
    help_text = "Comma Separated list of non-ticker words"


class FinancialSummarySectionName(StringPreference):
    section = fdm
    name = "financial_summary_section_name"
    default = "Financial Summary"

    verbose_name = "Financial Summary Section Name"
    help_text = "This name set the tab section name shown from the instance view"



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class InstrumentListThroughModelEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrumentlistthrough-list",
            args=[],
            request=self.request,
        )


class InstrumentListThroughModelInstrumentListEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrumentlist-instrumentlistthrough-list",
            args=[self.view.kwargs["instrument_list_id"]],
            request=self.request,
        )


class InstrumentListThroughModelInstrumentEndpointConfig(EndpointViewConfig):
    PK_FIELD = "instrument_list"  # we expect that users want to go directly to the Instrument List model

    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-instrumentlistthrough-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )

    def get_instance_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrumentlist-list",
            args=[],
            request=self.request,
        )



import pandas as pd
from django.utils.functional import cached_property
from wbcore.cache.decorators import cache_table
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf
from wbcore.serializers.fields.types import DisplayMode
from wbcore.utils.date import get_next_day_timedelta

from wbfdm.enums import DataType, PeriodType, StatementType
from wbfdm.filters import StatementFilter
from wbfdm.models.instruments import Instrument
from wbfdm.utils import rename_period_index_level_to_repr
from wbfdm.viewsets.configs.display.statements import StatementDisplayViewConfig
from wbfdm.viewsets.configs.endpoints.statements import StatementsEndpointViewConfig
from wbfdm.viewsets.configs.titles.statement_with_estimates import (
    StatementTitleViewConfig,
)

from ..mixins import InstrumentMixin


@cache_table(
    timeout=lambda view: get_next_day_timedelta(),
    key_prefix=lambda view: f"_{view.instrument.id}_{view.financial_analysis_key}_{view.data_type_key}",
)
class StatementPandasViewSet(InstrumentMixin, ExportPandasAPIViewSet):
    queryset = Instrument.objects.none()
    display_config_class = StatementDisplayViewConfig
    endpoint_config_class = StatementsEndpointViewConfig
    title_config_class = StatementTitleViewConfig
    filterset_class = StatementFilter

    financial_analysis_mapping = {
        "income-statement": (StatementType.INCOME_STATEMENT, "Income Statement"),
        "balance-sheet": (StatementType.BALANCE_SHEET, "Balance Sheet"),
        "cash-flow-statement": (StatementType.CASHFLOW_STATEMENT, "Cashflow Statement"),
    }

    @cached_property
    def data_type_key(self) -> str:
        return self.request.GET.get("data_type", "standardized")

    @cached_property
    def financial_analysis_key(self) -> str:
        return self.kwargs.get("statement", "income-statement")

    def get_queryset(self):
        return Instrument.objects.filter(id=self.instrument.id)

    def get_pandas_fields(self, request):
        return pf.PandasFields(
            fields=[
                pf.PKField(key="external_ordering", label="ID"),
                pf.CharField(key="external_code", label="Code"),
                pf.CharField(key="external_description", label="Description"),
                pf.SparklineField(key="progress", label="Yearly Trend", dimension="double"),
                *[
                    pf.FloatField(key=field, label=field, display_mode=DisplayMode.SHORTENED)
                    for field in self.columns
                    if field not in ["external_ordering", "external_code", "external_description", "progress"]
                ],
            ]
        )

    def get_dataframe(self, request, queryset, **kwargs):
        financial_analysis = self.financial_analysis_mapping.get(self.financial_analysis_key)[0]
        df = pd.DataFrame(
            Instrument.objects.filter(id=self.instrument.id).dl.statements(
                statement_type=financial_analysis,
                data_type=DataType(self.data_type_key),
                period_type=PeriodType.ALL,
            )
        )
        if not df.empty:
            df = df.pivot_table(
                index=["year", "interim", "period_type"],
                columns=["external_ordering", "external_code", "external_description"],
                values=["value"],
                aggfunc="first",
            )

            df = rename_period_index_level_to_repr(df)
            df = df.set_index([[f"{index[0]}-{index[1]}" for index in df.index]])
            df = df.T.reset_index()
        return df

    def manipulate_dataframe(self, df):
        if not df.empty:
            if year_cols := [col for col in df.columns if isinstance(col, str) and "Y" in col]:
                df["progress"] = df[year_cols].fillna(0).apply(lambda x: list(x.to_dict().items()), axis=1)
        return df

    @cached_property
    def columns(self):
        return self.df.columns

    @cached_property
    def year_columns(self):
        year_columns = list(map(lambda x: int(x.replace("-Y", "")), filter(lambda col: "Y" in col, self.columns)))
        year_columns.sort(reverse=True)
        return year_columns

    @cached_property
    def interim_columns(self):
        return list(filter(lambda col: "Y" not in col, self.columns))



from wbcore.menus import ItemPermission, MenuItem

ExchangeMenuItem = MenuItem(
    label="Exchange",
    endpoint="wbfdm:exchange-list",
    permission=ItemPermission(permissions=["wbfdm.view_exchange"], method=lambda request: request.user.is_internal),
)



from datetime import date

from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import DateRangeField, RangeOperators
from django.db import models, transaction
from django.utils.module_loading import import_string
from psycopg.types.range import DateRange
from slugify import slugify
from wbcore.contrib.io.mixins import ImportMixin
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.models import WBModel
from wbcore.utils.models import ComplexToStringMixin

from wbfdm.import_export.handlers.instrument_list import InstrumentListImportHandler
from wbfdm.models.instruments.instruments import Instrument


class InstrumentListThroughModel(ImportMixin, ComplexToStringMixin):
    import_export_handler_class = InstrumentListImportHandler
    """
    This model is not a Through model from a programming point of view, however it allows to link instrument list to
    instruments.
    """

    instrument_str = models.CharField(max_length=256)
    instrument = models.ForeignKey(
        to="wbfdm.Instrument",
        null=True,
        blank=True,
        limit_choices_to=models.Q(is_security=True),
        on_delete=models.SET_NULL,
    )
    instrument_list = models.ForeignKey(
        to="wbfdm.InstrumentList",
        on_delete=models.CASCADE,
    )
    timespan = DateRangeField(verbose_name="Timespan")
    comment = models.TextField(default="", blank=True)
    validated = models.BooleanField(default=False)

    def save(self, *args, **kwargs):
        if not self.timespan:
            self.timespan = DateRange(date.today(), None)
        super().save(*args, **kwargs)

    def compute_str(self) -> str:
        """
        Method to compute the string representation of the instance. It will save the string value to the computed_str
        field.

        Returns:
            The string representation of the instance.
        """
        if self.instrument and self.instrument.name_repr:
            return f"{self.instrument.name_repr} - {self.instrument_list.name}"
        return f"{self.instrument_str} - {self.instrument_list.name}"

    class Meta:
        verbose_name = "Instrument in Instrument List"
        constraints = [
            ExclusionConstraint(
                name="exclude_overlapping_instrument_list",
                condition=models.Q(instrument__isnull=False),
                expressions=[
                    ("timespan", RangeOperators.OVERLAPS),
                    ("instrument", RangeOperators.EQUAL),
                    ("instrument_list", RangeOperators.EQUAL),
                ],
            ),
        ]
        notification_types = [
            create_notification_type(
                "wbfdm.instrument_list_add",
                "Instrument added to Instrument List",
                "A notification when an instrument gets added to a list.",
                True,
                True,
                True,
            ),
        ]


class InstrumentList(WBModel):
    class InstrumentListType(models.TextChoices):
        WATCH = "WATCH", "Watch List"
        EXCLUSION = "EXCLUSION", "Exclusion List"
        INCLUSION = "INCLUSION", "Inclusion List"

    name = models.CharField(max_length=255)
    identifier = models.CharField(max_length=255, unique=True, blank=True)
    instrument_list_type = models.CharField(max_length=32, choices=InstrumentListType.choices, null=True, blank=True)
    constituent_loader_path = models.CharField(
        max_length=512,
        null=True,
        blank=True,
        help_text="If set, load the Instrument queryset returned by this function and set them as the new instrument list constituent",
    )
    constituent_loader_kwargs = models.JSONField(default=dict, blank=True)

    def __str__(self):
        return self.name

    def get_instruments(self, val_date: date, only_validated: bool = False) -> models.QuerySet[Instrument]:
        """
        Returns a QuerySet of Instrument objects associated with the current instrument list.

        This property filters the Instrument objects based on the related InstrumentListThroughModel
        and returns only those Instruments where the foreign key is not null.

        Returns:
            models.QuerySet[Instrument]: A QuerySet of Instrument objects.
        """
        rels = InstrumentListThroughModel.objects.filter(
            models.Q(instrument_list=self) & models.Q(instrument__isnull=False) & models.Q(timespan__contains=val_date)
        )
        if only_validated:
            rels = rels.filter(validated=True)
        return Instrument.objects.filter(id__in=(rels.values("instrument")))

    def save(self, *args, **kwargs):
        if not self.identifier:
            self.identifier = slugify(f"{self.name}-{self.id}")
        super().save(*args, **kwargs)

    def load_constituent(self, val_date: date) -> list[Instrument]:
        if self.constituent_loader_path:
            return import_string(self.constituent_loader_path)(self, val_date, **self.constituent_loader_kwargs)
        return []

    def refresh_constituent_from_list(self, val_date: date, instruments: list[Instrument]):
        active_instruments = self.get_instruments(val_date)
        active_instrument_ids = set(active_instruments.values_list("id", flat=True))
        new_instrument_ids = {instr.id for instr in instruments}
        objs_to_close = []

        with transaction.atomic():
            for rel_to_close in InstrumentListThroughModel.objects.filter(
                instrument_list=self, instrument_id__in=active_instrument_ids - new_instrument_ids
            ):
                rel_to_close.timespan = DateRange(rel_to_close.timespan.lower, val_date)
                objs_to_close.append(rel_to_close)
            if objs_to_close:
                InstrumentListThroughModel.objects.bulk_update(objs_to_close, ["timespan"], batch_size=10000)

            objs_to_create = []
            for instr in instruments:
                if instr.id not in active_instrument_ids:
                    objs_to_create.append(
                        InstrumentListThroughModel(
                            timespan=DateRange(val_date, None), instrument=instr, instrument_list=self
                        )
                    )
            if objs_to_create:
                InstrumentListThroughModel.objects.bulk_create(objs_to_create, batch_size=10000)

    class Meta:
        verbose_name = "Instrument List"
        verbose_name_plural = "Instrument Lists"
        permissions = (("administrate_instrumentlist", "Can administrate Instrument List"),)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbfdm:instrumentlist"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbfdm:instrumentlistrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"



import random
from datetime import date
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
from faker import Faker

from wbfdm.analysis.financial_analysis.utils import FinancialAnalysisResult, Loader
from wbfdm.dataloaders.proxies import InstrumentDataloaderProxy
from wbfdm.enums import Financial, MarketData

fake = Faker()


@pytest.mark.django_db
class TestLoader:
    @pytest.fixture
    def data(self, value, year):
        """
        Data Fixture for the Loader class
        """
        year = int(year)
        data = []
        fake_key = fake.word()
        for year_delta in [0, 1]:
            for interim in range(1, 5):
                data.append(
                    {
                        "year": year + year_delta,
                        "interim": interim,
                        "period_type": "Q",
                        "period_end_date": date(year, interim * 3, 1),
                        "estimate": fake.pybool(),
                        "source": "dsws",
                        "financial": value,
                        "value": fake.pyfloat(),
                        fake_key: fake.word(),  # add noise key
                    }
                )
        # add a uncomplete year for year - 2
        data.append(
            {
                "year": year - 1,
                "interim": 1,
                "period_type": "Q",
                "period_end_date": date(year, 3, 1),
                "estimate": fake.pybool(),
                "source": "dsws",
                "financial": value,
                "value": fake.pyfloat(),
                fake_key: fake.word(),  # add noise key
            }
        )
        return data

    @pytest.fixture()
    def loader(self, instrument, value):
        return Loader(instrument, [Financial(value)])

    @pytest.mark.parametrize("value, year", [(random.choice(Financial.values()), fake.year())])
    @patch.object(InstrumentDataloaderProxy, "financials")
    def test_load(self, mock_fct, value, year, data, loader):
        # Test that the load method returnn the normalized dataframe

        mock_fct.return_value = data
        pd.testing.assert_frame_equal(loader.load(), loader._normalize_df(*loader._get_base_df()), check_exact=True)

    @pytest.mark.parametrize("value, year", [(random.choice(Financial.values()), fake.year())])
    @patch.object(InstrumentDataloaderProxy, "financials")
    def test_load_no_data(self, mock_fct, value, year, data, loader):
        # Test that the load method returnn the normalized dataframe

        mock_fct.return_value = []
        assert loader.load().empty

    @pytest.mark.parametrize("value, year", [(random.choice(Financial.values()), fake.year())])
    @patch.object(InstrumentDataloaderProxy, "financials")
    def test__get_base_df(self, mock_fct, value, year, data, loader):
        year = int(year)
        mock_fct.return_value = data
        df, source_df = loader._get_base_df()
        time_idx = df.index.droplevel([3, 4])
        # Test if the _get_base_df private method returned tuple have the same index
        pd.testing.assert_index_equal(time_idx, source_df.index)

        # Test if the _get_base_df private method returned dataframe index corresponds to the expected year and interim
        assert time_idx.tolist() == [
            (year - 1, 1, "Q"),
            (year, 1, "Q"),
            (year, 2, "Q"),
            (year, 3, "Q"),
            (year, 4, "Q"),
            (year + 1, 1, "Q"),
            (year + 1, 2, "Q"),
            (year + 1, 3, "Q"),
            (year + 1, 4, "Q"),
        ]
        assert source_df.unique() == "dsws"
        assert df.columns == [value]

    @pytest.mark.parametrize("market_value", [random.choice(MarketData.values())])
    @patch.object(InstrumentDataloaderProxy, "market_data")
    def test__annotate_market_data(self, mock_fct, instrument, market_value):
        # test if annotation market data works and are merged properly to the initial dataframe
        d1 = date(2021, 3, 30)
        d2 = date(2023, 3, 31)
        p1 = fake.pyfloat()
        p2 = fake.pyfloat()

        df = pd.DataFrame(
            [
                {
                    "year": d1.year,
                    "estimate": False,
                    "interim": 1,
                    "period_type": "P",
                    "period_end_date": d1,
                    "financial": "financial",
                    "value": fake.pyfloat(),
                },
                {
                    "year": d2.year,
                    "estimate": False,
                    "interim": 1,
                    "period_type": "P",
                    "period_end_date": d2,
                    "financial": "financial",
                    "value": fake.pyfloat(),
                },
            ]
        )
        df["period_end_date"] = pd.to_datetime(df["period_end_date"])
        df = df.pivot_table(
            index=["year", "interim", "period_type", "estimate", "period_end_date"],
            columns="financial",
            values="value",
        )
        mock_fct.return_value = [{"valuation_date": d1, market_value: p1}, {"valuation_date": d2, market_value: p2}]
        res = Loader(instrument, [])._annotate_market_data(df, [MarketData(market_value)])
        assert res.loc[(d1.year, 1, "P", False, "2021-03-30"), market_value] == p1
        assert res.loc[(d2.year, 1, "P", False, "2023-03-31"), market_value] == p2

    @pytest.mark.parametrize("statement_value", [random.choice(Financial.values())])
    @patch.object(InstrumentDataloaderProxy, "statements")
    def test__annotate_statement_data(self, mock_fct, instrument, statement_value):
        # test if annotation statement works and are merged properly to the initial dataframe

        d1 = date(2021, 3, 30)
        d2 = date(2023, 3, 31)
        value = fake.pyfloat()

        df = pd.DataFrame(
            [
                {
                    "year": d1.year,
                    "estimate": False,
                    "interim": 1,
                    "period_type": "P",
                    "period_end_date": d1,
                    "financial": "base",
                    "value": fake.pyfloat(),
                },
                {
                    "year": d2.year,
                    "estimate": False,
                    "interim": 1,
                    "period_type": "P",
                    "period_end_date": d2,
                    "financial": "base",
                    "value": fake.pyfloat(),
                },
            ]
        )
        df["period_end_date"] = pd.to_datetime(df["period_end_date"])
        df = df.pivot_table(
            index=["year", "interim", "period_type", "estimate", "period_end_date"],
            columns="financial",
            values="value",
        )
        mock_fct.return_value = [
            {
                "year": d1.year,
                "estimate": False,
                "interim": 1,
                "period_type": "P",
                "period_end_date": d1,
                "financial": statement_value,
                "value": value,
            }
        ]

        res = Loader(instrument, [])._annotate_statement_data(df, [Financial(statement_value)])
        assert res.loc[(d1.year, 1, "P", False, "2021-03-30"), statement_value] == value
        assert np.isnan(res.loc[(d2.year, 1, "P", False, "2023-03-31"), statement_value])

    def test__normalize_df(self, instrument):
        d1 = date(2021, 3, 30)
        df = pd.DataFrame(
            [
                {
                    "source": "dsws",
                    "year": d1.year,
                    "estimate": False,
                    "interim": 1,
                    "period_end_date": d1,
                    "period_type": "S",
                    "financial": "base",
                    "value": fake.pyfloat(),
                },
                {
                    "source": "dsws",
                    "year": d1.year,
                    "estimate": False,
                    "interim": 2,
                    "period_end_date": d1,
                    "period_type": "S",
                    "financial": "base",
                    "value": fake.pyfloat(),
                },
            ]
        )
        source_df = df.source
        df["period_end_date"] = pd.to_datetime(df["period_end_date"])
        df = df.pivot_table(
            index=["year", "interim", "period_type", "estimate", "period_end_date"],
            columns="financial",
            values="value",
        )
        res = Loader(instrument, [])._normalize_df(df, source_df)
        # Test if normalize appends the missing "yearly" row and convert the interim into their verbose representation
        assert res.index.tolist() == [
            (d1.year, "Y"),
            (d1.year, "S1"),
            (d1.year, "S2"),
        ]  # We test that the missing yearly row is reindexed
        assert np.isnan(res.loc[(d1.year, "Y"), "base"])  # check that the extrapolated data is nan

    def test__normalize_df_with_duplicates(self, instrument):
        # we test if normalize detects the duplicates, remove it by taking the first and add the interim into the errors list
        d1 = date(2021, 3, 30)
        v1 = fake.pyfloat()
        v2 = fake.pyfloat()
        df = pd.DataFrame(
            [
                {
                    "source": "dsws",
                    "year": d1.year,
                    "estimate": False,
                    "interim": 1,
                    "period_end_date": d1,
                    "financial": "base",
                    "period_type": "P",
                    "value": v1,
                },
                {
                    "source": "dsws",
                    "year": d1.year,
                    "estimate": False,
                    "interim": 1,
                    "period_end_date": date(2021, 1, 30),
                    "financial": "base",
                    "period_type": "P",
                    "value": v2,
                },
            ]
        )
        source_df = df[["year", "interim", "source"]].groupby(["year", "interim"]).first().source
        df["period_end_date"] = pd.to_datetime(df["period_end_date"])
        df = df.pivot_table(
            index=["year", "interim", "period_type", "estimate", "period_end_date"],
            columns="financial",
            values="value",
        )
        loader = Loader(instrument, [])
        res = loader._normalize_df(df, source_df)
        assert res.index.tolist() == [(d1.year, "Y"), (d1.year, "P1")]
        assert (
            res.loc[(d1.year, "P1"), "base"] == v2
        )  # check that the retained value is the first one (with ordered index)
        assert loader.errors["duplicated_interims"] == [f"{d1.year} Interim P1"]


class TestFinancialAnalysisResult:
    @pytest.fixture()
    def df(self, year, value):
        df = pd.DataFrame(
            [
                {
                    "year": year,
                    "estimate": False,
                    "interim": "P1",
                    "period_end_date": date(int(year), 3, 1),
                    "financial": value,
                    "value": fake.pyfloat(),
                },
                {
                    "year": year,
                    "estimate": True,
                    "interim": "Y",
                    "period_end_date": date(int(year), 3, 1),
                    "financial": value,
                    "value": fake.pyfloat(),
                },
            ]
        )
        df = df.pivot_table(
            index=["year", "interim", "estimate", "period_end_date"], columns="financial", values="value"
        )
        return df.reset_index(level=[2, 3])

    @pytest.mark.parametrize("value, year", [(random.choice(Financial.values()), fake.year())])
    def test_class(self, value, year, df):
        """
        Test the basic esthetic transform (Transposing + group key addition + convertion of multi index to representation
        """
        res = FinancialAnalysisResult(df)
        assert res.formatted_df.loc[0, f"{year}-P1"] == df.loc[(year, "P1"), value]
        assert res.formatted_df.loc[0, f"{year}-Y"] == df.loc[(year, "Y"), value]
        assert res.formatted_df.loc[0, "financial"] == Financial.name_mapping()[value]
        assert res.formatted_df.loc[0, "_group_key"] == value
        assert res.estimated_mapping == {f"{year}-P1": False, f"{year}-Y": True}



from datetime import date

from celery import shared_task
from django.contrib.contenttypes.models import ContentType
from wbcore.workers import Queue

from wbfdm.contrib.metric.dispatch import compute_metrics


@shared_task(queue=Queue.EXTENDED_BACKGROUND.value)
def compute_metrics_as_task(
    val_date: date | None = None,
    key: str | None = None,
    basket_content_type_id: int | None = None,
    basket_id: int | None = None,
    **kwargs,
):
    """
    Run the orchestrator as a async task periodically for all keys and baskets

    Args:
        val_date (date): Compute the metric for the given date (Default to None). If None, let the metric backends decide what is the last valide date to use

    """
    basket = None
    if basket_content_type_id and basket_id:
        basket = ContentType.objects.get(id=basket_content_type_id).get_object_for_this_type(pk=basket_id)
    compute_metrics(val_date=val_date, key=key, basket=basket, **kwargs)



import logging
from contextlib import suppress

import numpy as np
from django.conf import settings
from django.contrib.messages import warning
from django.utils.functional import cached_property
from wbcore.cache.decorators import cache_table
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf
from wbcore.serializers.fields.types import DisplayMode
from wbcore.utils.date import get_next_day_timedelta

from wbfdm.analysis.financial_analysis.statement_with_estimates import (
    StatementWithEstimates,
)
from wbfdm.enums import CalendarType
from wbfdm.filters import StatementWithEstimateFilter
from wbfdm.models.instruments import Instrument
from wbfdm.viewsets.configs.display.statement_with_estimates import (
    StatementWithEstimatesDisplayViewConfig,
)
from wbfdm.viewsets.configs.endpoints.statements import StatementsEndpointViewConfig
from wbfdm.viewsets.configs.titles.statement_with_estimates import (
    StatementTitleViewConfig,
)

from ..mixins import InstrumentMixin

logger = logging.getLogger("pms")


@cache_table(
    timeout=lambda view: get_next_day_timedelta(),
    key_prefix=lambda view: f"_{view.instrument.id}_{view.financial_analysis_key}_{view.calendar_type}",
)
class StatementWithEstimatesPandasViewSet(InstrumentMixin, ExportPandasAPIViewSet):
    queryset = Instrument.objects.none()
    display_config_class = StatementWithEstimatesDisplayViewConfig
    endpoint_config_class = StatementsEndpointViewConfig
    title_config_class = StatementTitleViewConfig
    filterset_class = StatementWithEstimateFilter

    financial_analysis_mapping = {
        "income": ("income_statement_with_estimate", "Income Statement"),
        "balancesheet": ("balance_sheet_with_estimate", "Balance Sheet"),
        "cashflow": ("cash_flow_statement_with_estimate", "Cash Flow Statement"),
        "ratios": ("ratios_with_estimate", "Ratios"),
        "summary": ("summary_with_estimate", "Summary"),
        "margins": ("margins_with_estimates", "Margins"),
        "cashflow-ratios": ("cashflow_ratios_with_estimates", "Cashflow Ratios"),
        "asset-turnover-ratios": ("asset_turnover_with_estimates", "Asset-Turnover Ratios"),
        "credit": ("credit_with_estimates", "Credit"),
        "long-term-solvency": ("long_term_solvency_with_estimates", "Long-term Solvency"),
        "short-term-liquidity": ("short_term_liquidity_with_estimates", "Short-term Liquidity"),
    }

    def get_queryset(self):
        return Instrument.objects.filter(id=self.instrument.id)

    def get_pandas_fields(self, request):
        return pf.PandasFields(
            fields=[
                *[pf.FloatField(key=field, label=field, display_mode=DisplayMode.SHORTENED) for field in self.columns],
                pf.PKField(key="id", label="ID"),
                pf.CharField(key="financial", label="Financial"),
                pf.CharField(key="_group_key", label="Group Key"),
                pf.JsonField(key="_overwrites", label="Overwrites"),
                pf.SparklineField(key="progress", label="Yearly Trend", dimension="double"),
            ]
        )

    def get_dataframe(self, request, queryset, **kwargs):
        statement_with_estimate = StatementWithEstimates(self.instrument, calendar_type=self.calendar_type)
        financial_analysis_result = getattr(
            statement_with_estimate, self.financial_analysis_mapping[self.financial_analysis_key][0]
        )
        df = financial_analysis_result.formatted_df

        year_columns = list(filter(lambda col: "Y" in col, df.columns))
        if year_columns:
            df["progress"] = (
                df[year_columns]
                .replace([np.inf, -np.inf, np.nan], None)
                .apply(lambda x: list(x.to_dict().items()), axis=1)
            )

        self.extra_cache_kwargs["_estimate_mapping"] = financial_analysis_result.estimated_mapping
        self.extra_cache_kwargs["_columns"] = df.columns
        self.extra_cache_kwargs["_errors"] = financial_analysis_result.errors

        if financial_analysis_result.errors.get("duplicated_interims", []) and not settings.DEBUG:
            with suppress(ModuleNotFoundError):
                logger.warning("Finanical Statement: Duplicate index detected.", extra={"instrument": self.instrument})
        return df

    def add_messages(self, request, instance=None, **kwargs):
        if self.errors:
            message = """
            <p>While gathering the financial data, we detected the following issues from the data vendor:</p>
            <ul>
            """
            if duplicated_interim := self.errors.get("duplicated_interim", []):
                message += f"""
                <li>Duplicated data for for interim period <strong>{", ".join(duplicated_interim)}</strong>
                    <ul>
                    <li>First available data was used and the rest was ignored
                    </li>
                    </ul>
                </li>
                """
            if missing_data := self.errors.get("missing_data", []):
                for error in missing_data:
                    message += f"<li>{error}</li>"
            message += "</ul>"
            warning(request, message, extra_tags="auto_close=0")

    # Cached attributes as properties
    @cached_property
    def calendar_type(self):
        return CalendarType(self.request.GET.get("calendar_type", CalendarType.FISCAL.value))

    @cached_property
    def financial_analysis_key(self):
        return self.kwargs.get("statement", "income")

    @cached_property
    def estimate_mapping(self):
        return getattr(self, "_estimate_mapping", {})

    @cached_property
    def errors(self):
        return getattr(self, "_errors", {})

    @cached_property
    def columns(self):
        return self.df.columns

    def get_ordering_fields(self):
        return [x for x in self.columns if x != "progress"]

    @cached_property
    def year_columns(self):
        return list(filter(lambda col: "Y" in col, self.columns))

    @cached_property
    def interim_columns(self):
        return list(filter(lambda col: "Y" not in col, self.columns))



from datetime import date

import pandas as pd

from wbfdm.enums import Financial, PeriodType, SeriesType
from wbfdm.models import Instrument
from wbfdm.utils import rename_period_index_level_to_repr


def financial_metric_estimate_analysis(instrument_id: int, financial: Financial) -> tuple[pd.DataFrame, dict, list]:
    estimate_mapping = {}
    columns = []
    if not (
        df := pd.DataFrame(
            Instrument.objects.filter(id=instrument_id).dl.financials(
                values=[financial],
                series_type=SeriesType.FULL_ESTIMATE,
                period_type=PeriodType.ALL,
                from_year=date.today().year - 5,
            ),
        )
    ).empty:
        df = df.pivot_table(
            index=["year", "interim", "period_type", "estimate"],
            columns=["financial"],
            values=["value", "difference_pct", "value_high", "value_low", "value_amount", "value_stdev"],
            dropna=False,
        )
        df["difference_pct"] = df["difference_pct"] * 100
        df = rename_period_index_level_to_repr(df)
        df = df.set_index([[f"{index[0]}-{index[1]}" for index in df.index]])

        columns = list(df.index)

        df = df[["value", "difference_pct", "value_high", "value_low", "value_stdev", "value_amount"]]
        df = df.rename(
            columns={
                "difference_pct": "Diff (%)",
                "value": "Estimate",
                "value_amount": "#",
                "value_high": "High Est.",
                "value_low": "Low Est.",
                "value_stdev": "St.Dev. Est.",
            }
        )
        df = df.T

        df["financial"] = [index[0] for index in df.index]
        df = df.set_index([[f"{index[0]}-{index[1]}" for index in df.index]])
        df = df.reset_index()

    return df, estimate_mapping, columns


def financial_metric_growths(instrument_id: int, financial: Financial):
    # Get the base dataframe from a dataloader with the given financial
    if not (
        df := pd.DataFrame(
            Instrument.objects.filter(id=instrument_id).dl.financials(
                values=[financial],
                series_type=SeriesType.COMPLETE,
                period_type=PeriodType.ALL,
                from_year=date.today().year - 5,
            )
        )
    ).empty:
        # Pivot the dataframe to get the financial in the correct format
        df = df.pivot_table(index=["year", "interim", "period_type"], columns="financial", values="value")

        # Compute the growth factors
        df[f"{financial.value}_qq"] = df.loc[df.index.get_level_values("interim") != 0].pct_change() * 100
        df[f"{financial.value}_yy_y"] = (
            df.loc[df.index.get_level_values("interim") == 0, financial.value].pct_change() * 100
        )
        df[f"{financial.value}_yy_q"] = (
            df.loc[df.index.get_level_values("interim") != 0, financial.value].pct_change(4) * 100
        )
        df[f"{financial.value}_yy"] = df[f"{financial.value}_yy_y"].combine_first(df[f"{financial.value}_yy_q"])

        # Select on the two growth columns
        df = df[[f"{financial.value}_yy", f"{financial.value}_qq"]]

        df = rename_period_index_level_to_repr(df)
        df = df.set_index([[f"{index[0]}-{index[1]}" for index in df.index]])
        df = df.rename(columns={f"{financial.value}_yy": "YoY Growth (%)", f"{financial.value}_qq": "QoQ Growth (%)"})
        df = df.T
        df = df.reset_index().reset_index()

    return df



import operator
from contextlib import suppress
from datetime import datetime
from functools import reduce
from typing import Any, Dict, Optional

from django.contrib.postgres.search import TrigramSimilarity
from django.core.exceptions import MultipleObjectsReturned
from django.db import IntegrityError, models
from django.db.models import Exists, OuterRef, Q
from slugify import slugify
from wbcore.contrib.currency.import_export.handlers import CurrencyImportHandler
from wbcore.contrib.geography.models import Geography
from wbcore.contrib.io.exceptions import DeserializationError
from wbcore.contrib.io.imports import ImportExportHandler, ImportState

from wbfdm.models.exchanges import Exchange


class InstrumentLookup:
    ORDERED_KEYS = [
        "instrument_type",
        "isin",
        "refinitiv_identifier_code",
        "refinitiv_mnemonic_code",
        "identifier",
        "ticker",
        "currency",
    ]

    def __init__(self, model, trigram_similarity_min_score: float = 0.8):
        self.cache = {}
        self.model = model
        self.trigram_similarity_min_score = trigram_similarity_min_score

    @classmethod
    def _get_cache_key(cls, **data):
        return "-".join([f"{k}:{slugify(str(data[k]))}" for k in cls.ORDERED_KEYS if data.get(k, None) is not None])

    def _get_cache(self, **kwargs):
        cache_key = self._get_cache_key(**kwargs)
        if cache_key and cache_key in self.cache:
            return self.cache[cache_key]

    def _set_cache(self, instrument, **kwargs):
        cache_key = self._get_cache_key(**kwargs)
        self.cache[cache_key] = instrument

    def _get_instruments_from_identifiers(self, queryset, **identifiers):
        # Try exact lookup on the filtered out universe
        identifiers = {k: v for k, v in identifiers.items() if v is not None}
        for identifier_key in [
            "isin",
            "refinitiv_identifier_code",
            "refinitiv_mnemonic_code",
            "sedol",
            "cusip",
            "identifier",
        ]:
            if identifier := identifiers.get(identifier_key, None):
                with suppress(self.model.DoesNotExist, MultipleObjectsReturned):
                    identifier = str(identifier)
                    if (
                        identifier_key != "refinitiv_identifier_code"
                    ):  # RIC cannot be uppercased because its symbology implies meaning for lowercase characters
                        identifier = identifier.upper()
                    # we try to lookup the instrument with the unique identifier
                    instrument = queryset.get(**{identifier_key: identifier})
                    return self.model.objects.filter(id=instrument.id)
        return queryset

    def _filter_identifiers(self, queryset, **identifiers):
        lookup_fields = [
            "isin",
            "refinitiv_identifier_code",
            "refinitiv_mnemonic_code",
            "refinitiv_ticker",
            "identifier",
            "ticker",
        ]
        conditions = []
        for field in lookup_fields:
            if field_value := identifiers.get(field, None):
                conditions.append(Q(**{f"{field}": field_value}))
                if field == "isin":
                    conditions.append(Q(old_isins__contains=[field_value]))
                if field == "ticker":
                    conditions.append(
                        Q(ticker__regex=rf"^{field_value}(\.[A-Za-z])$")
                    )  # initial regex: ([A-Za-z]?|\.?[A-Za-z])$ This led to a false positive, I don't remember why we allow an extra random character to be consider for eval
                    conditions.append(Q(refinitiv_mnemonic_code=f"@{field_value}"))

        if conditions:
            return queryset.filter(reduce(operator.or_, conditions))
        return queryset

    def _filter_name(self, queryset, name):
        instruments = queryset.annotate(similarity_score=TrigramSimilarity("name", name))
        if (
            queryset.count() > 1
            and instruments.filter(similarity_score__gt=self.trigram_similarity_min_score).count() == 1
        ):
            return instruments.filter(similarity_score__gt=self.trigram_similarity_min_score)
        return queryset

    def _filter_currency(self, queryset, currency):
        qs = queryset.filter(currency=currency)
        if qs.exists():
            return qs
        return queryset

    def _filter_exchange(self, queryset, exchange):
        return queryset.filter(exchange=exchange)

    def _filter_instrument_type(self, queryset, instrument_type):
        if not isinstance(instrument_type, str):
            instrument_type = instrument_type.key
        if instrument_type == "equity":
            qs = queryset.filter(instrument_type__key__in=["equity", "american_depository_receipt"])
        else:
            qs = queryset.filter(instrument_type__key=instrument_type)
        if qs.exists():
            return qs
        return queryset

    def _get_queryset(self, only_investable_universe: bool = False):
        instruments = self.model.objects.filter(is_security=True)
        if only_investable_universe:
            instruments = instruments.annotate(
                has_children_in_investable_universe=Exists(
                    self.model.objects.filter(is_investable_universe=True, parent=OuterRef("id"))
                )
            ).filter(Q(is_investable_universe=True) | Q(has_children_in_investable_universe=True))
        return instruments

    def _get_security_candidates(  # noqa: C901
        self,
        queryset,
        instrument_type=None,
        currency=None,
        name=None,
        **identifiers,
    ):
        # We need to lookup ticker because some provider gives us ticker with or without space in it
        instruments = self._get_instruments_from_identifiers(queryset, **identifiers)
        if instrument_type:
            instruments = self._filter_instrument_type(instruments, instrument_type)

        if instruments.count() == 1:
            return instruments

        instruments = self._filter_identifiers(instruments, **identifiers)

        if currency:
            instruments = self._filter_currency(instruments, currency)
        if name:
            instruments = self._filter_name(instruments, name)
        return instruments

    def _lookup_quote(
        self,
        security,
        currency=None,
        exchange=None,
        refinitiv_identifier_code=None,
        refinitiv_mnemonic_code=None,
        **kwargs,
    ):
        quotes = security.children.all()
        if not quotes.exists():
            return security
        if quotes.count() == 1:
            return quotes.first()
        if exchange:
            quotes = self._filter_exchange(quotes, exchange)
        if currency:
            quotes = self._filter_currency(quotes, currency)
        if refinitiv_mnemonic_code or refinitiv_identifier_code:
            quotes = self._filter_identifiers(
                quotes,
                refinitiv_identifier_code=refinitiv_identifier_code,
                refinitiv_mnemonic_code=refinitiv_mnemonic_code,
            )
        if quotes.filter(is_investable_universe=True).count() == 1:
            quotes = quotes.filter(is_investable_universe=True)
        if quotes.count() == 1:
            return quotes.first()
        return security.children.filter(is_primary=True).first()

    def lookup_security(self, **kwargs):
        queryset = self._get_queryset(only_investable_universe=False)
        security_candidates = self._get_security_candidates(queryset, **kwargs)

        if security_candidates.count() == 1:
            return security_candidates.first()
        # queryset = self._get_queryset(only_investable_universe=False)
        # security_candidates = self._get_security_candidates(queryset, **kwargs)

        if security_candidates.exists() > 1:
            security_candidates = security_candidates.filter(delisted_date__isnull=True)
        if security_candidates.count() == 1:
            return security_candidates.first()

    def lookup(self, only_security: bool = False, **lookup_kwargs):
        # To speed up lookup process, we try to get the quote from the investable universe first
        if instrument := self._get_cache(**lookup_kwargs):
            return instrument
        instrument = self.lookup_security(**lookup_kwargs)
        if not only_security and instrument:
            instrument = self._lookup_quote(instrument, **lookup_kwargs)
        if instrument:
            self._set_cache(instrument, **lookup_kwargs)
        return instrument


class InstrumentImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "wbfdm.Instrument"
    allow_update_save_failure = True
    exclude_update_fields = ["name", "isin", "country"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_lookup = InstrumentLookup(self.model)
        self.currency_handler = CurrencyImportHandler(self.import_source)

    def _deserialize(self, data: Dict[str, Any]):
        from wbfdm.models import Classification, InstrumentType

        if isinstance(data, int):
            data = dict(id=data)
        if data.get("currency", None):
            data["currency"] = self.currency_handler.process_object(
                data["currency"], read_only=True, raise_exception=False
            )[0]
        if instrument_type := data.get("instrument_type", None):
            if isinstance(instrument_type, str):
                data["instrument_type"] = InstrumentType.objects.get_or_create(
                    key=instrument_type,
                    defaults={"name": instrument_type.title(), "short_name": instrument_type.title()},
                )[0]
            elif isinstance(instrument_type, int):
                data["instrument_type"] = InstrumentType.objects.get(id=instrument_type)
        if data.get("country", None):
            data["country"] = Geography.dict_to_model(data["country"])
        if data.get("headquarter_city", None):
            data["headquarter_city"] = Geography.dict_to_model(data["headquarter_city"], level=Geography.Level.CITY)
        if inception_date := data.get("inception_date", None):
            data["inception_date"] = datetime.strptime(inception_date, "%Y-%m-%d").date()
        if classifications := data.pop("classifications", None):
            data["classifications"] = [Classification.dict_to_model(c) for c in classifications]
        if (exchange_data := data.pop("exchange", None)) and isinstance(exchange_data, dict):
            sanitized_dict = {k: v for k, v in exchange_data.items() if v is not None}
            if sanitized_dict:
                data["exchange"] = Exchange.dict_to_model(sanitized_dict)

    def _get_instance(
        self,
        data: Dict[str, Any],
        history: Optional[models.QuerySet] = None,
        only_security: bool = False,
        **kwargs,
    ) -> models.Model:
        if isinstance(data, self.model):
            return data
        if instrument_id := data.pop("id", None):
            try:
                return self.model.objects.get(id=instrument_id)
            except self.model.DoesNotExist as e:
                raise DeserializationError("Instrument id does not match an existing instrument") from e
        else:
            return self.instrument_lookup.lookup(only_security=only_security, **data)

    def _create_instance(self, data: Dict[str, Any], **kwargs) -> models.Model:
        from wbfdm.models.instruments.instruments import Instrument

        if not data.get("name", None):
            raise DeserializationError("Can't create an instrument without at least a name")
        classifications = data.pop("classifications", None)
        try:
            obj = Instrument.objects.create(
                **data,
                import_source=self.import_source,
            )
            if classifications:
                obj.classifications.set([c for c in classifications if c])
            return obj
        except IntegrityError:
            self.import_source.log += f"\nError while creating new instrument with data {data}"

        return None

    def process_object(self, underlying_instrument_data, **kwargs):
        if underlying_instrument_data:
            if isinstance(underlying_instrument_data, self.model):
                return underlying_instrument_data, ImportState.UNMODIFIED
            if isinstance(underlying_instrument_data, int):
                underlying_instrument_data = dict(id=underlying_instrument_data)
        return super().process_object(
            underlying_instrument_data,
            include_update_fields=[
                "isin",
                "ticker",
                "refinitiv_mnemonic_code",
                "refinitiv_identifier_code",
                "country",
                "exchange",
            ],
            **kwargs,
        )



from typing import Optional

from django.utils.translation import gettext, pgettext_lazy
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig


class InstrumentPriceDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="valuation_date", label=gettext("Date")),
                dp.Field(key="open", label=pgettext_lazy("financial", "Open")),
                dp.Field(key="high", label=pgettext_lazy("financial", "High")),
                dp.Field(key="low", label=pgettext_lazy("financial", "Low")),
                dp.Field(key="close", label=pgettext_lazy("financial", "Close")),
                dp.Field(key="volume", label=pgettext_lazy("financial", "Volume")),
                dp.Field(key="outstanding_shares", label=pgettext_lazy("financial", "Oustanding Shares")),
                dp.Field(key="market_capitalization", label=pgettext_lazy("financial", "Market Capitalization")),
                dp.Field(
                    key="market_capitalization_consolidated",
                    label=pgettext_lazy("financial", "Market Capitalization (Consolidated)"),
                ),
            ],
        )



from contextlib import suppress

import pandas as pd
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf
from wbcore.contrib.permission.permissions import InternalUserPermissionMixin
from wbcore.utils.date import get_date_interval_from_request
from wbcore.utils.strings import format_number

from wbfdm.filters import MonthlyPerformancesInstrumentFilterSet
from wbfdm.models.instruments import Instrument
from wbfdm.viewsets.configs import (
    MonthlyPerformancesInstrumentDisplayViewConfig,
    MonthlyPerformancesInstrumentEndpointConfig,
    MonthlyPerformancesInstrumentTitleConfig,
)

from ..mixins import InstrumentMixin


class MonthlyPerformancesInstrumentPandasViewSet(InstrumentMixin, InternalUserPermissionMixin, ExportPandasAPIViewSet):
    IDENTIFIER = "wbfdm:monthlyreturn"

    display_config_class = MonthlyPerformancesInstrumentDisplayViewConfig
    title_config_class = MonthlyPerformancesInstrumentTitleConfig
    endpoint_config_class = MonthlyPerformancesInstrumentEndpointConfig
    filterset_class = MonthlyPerformancesInstrumentFilterSet

    pandas_fields = pf.PandasFields(
        fields=(
            pf.PKField(key="id", label="ID"),
            pf.YearField(key="year", label="Year"),
            pf.FloatField(key="1", label="January", precision=2, percent=True),
            pf.FloatField(key="2", label="February", precision=2, percent=True),
            pf.FloatField(key="3", label="March", precision=2, percent=True),
            pf.FloatField(key="4", label="April", precision=2, percent=True),
            pf.FloatField(key="5", label="May", precision=2, percent=True),
            pf.FloatField(key="6", label="June", precision=2, percent=True),
            pf.FloatField(key="7", label="July", precision=2, percent=True),
            pf.FloatField(key="8", label="August", precision=2, percent=True),
            pf.FloatField(key="9", label="September", precision=2, percent=True),
            pf.FloatField(key="10", label="October", precision=2, percent=True),
            pf.FloatField(key="11", label="November", precision=2, percent=True),
            pf.FloatField(key="12", label="December", precision=2, percent=True),
            pf.FloatField(key="annual", label="Yearly", precision=2, percent=True),
        )
    )

    queryset = Instrument.objects.all()
    ordering_fields = (
        "year",
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "10",
        "11",
        "annual",
    )
    ordering = ["-year"]

    def get_aggregates(self, request, df):
        if df.empty:
            return {}
        annual_perfs = df.annual + 1
        total_perf = annual_perfs.product(axis=0, skipna=False) - 1
        return {"annual": {"Σ": format_number(total_perf)}}

    def get_dataframe(self, request, queryset, **kwargs):
        d1, d2 = get_date_interval_from_request(request, exclude_weekend=True, date_range_fieldname="period")
        df = pd.DataFrame()
        if d1 and d2:
            df, _ = self.instrument.get_monthly_return_summary(d1, d2)
            if not df.empty:
                df = df.pivot_table(index="year", columns=["month"], values="performance").rename_axis(None, axis=1)
                if benchmark_id := request.GET.get("benchmark", None):
                    with suppress(Instrument.DoesNotExist):
                        benchmark = Instrument.objects.get(id=benchmark_id)
                        df_benchmark, _ = benchmark.get_monthly_return_summary(d1, d2)
                        df_benchmark = df_benchmark.pivot_table(
                            index="year", columns=["month"], values="performance"
                        ).rename_axis(None, axis=1)
                        df = df - df_benchmark

                df = df.reset_index()
                df["id"] = df.index
                df.columns = df.columns.astype(str)
        return df



from contextlib import suppress
from datetime import date, timedelta

from celery import shared_task
from django.db import transaction
from django.db.models import ProtectedError, Q
from pandas.tseries.offsets import BDay
from tqdm import tqdm
from wbcore.utils.cache import mapping
from wbcore.workers import Queue

from wbfdm.models import Instrument, InstrumentList, InstrumentPrice
from wbfdm.sync.runner import (  # noqa: F401
    initialize_exchanges,
    initialize_instruments,
    synchronize_exchanges,
    synchronize_instruments,
)

from .contrib.metric.signals import instrument_metric_updated
from .signals import investable_universe_updated


@shared_task(queue=Queue.BACKGROUND.value)
def update_of_investable_universe_data(
    start: date | None = None,
    end: date | None = None,
    with_background_tasks: bool = True,
    instrument_ids: list[int] | None = None,
):
    """
    Update the investable universe data on a daily basis.

    Parameters:
    - start (date | None): The start date for updating data. If None, defaults to three business days before 'end'.
    - end (date | None): The end date for updating data. If None, defaults to the current date.
    - with_background_tasks (bool | True): If true, will trigger the post import background tasks automatically.
    - instrument_ids (list[int] | None): if specified, narrow down the instrument queryset with this list ids

    Notes:
    - The function resets the investable universe by marking all instruments as not in the investable universe.
    - It then updates all instruments marked as part of the investable universe.
    - If 'end' is not provided, it defaults to the current date.
    - If 'start' is not provided, it defaults to three business days before 'end'.

    Returns:
    None
    """
    if not end:
        end = (
            date.today() - BDay(1)
        ).date()  # we don't import today price in case the dataloader returns duplicates (e.g. DSWS)
    if not start:
        start = (end - BDay(3)).date()  # override three last day by default
    Instrument.investable_universe.update(
        is_investable_universe=True
    )  # ensure all the investable universe is marked as such
    instruments = Instrument.active_objects.filter(is_investable_universe=True, delisted_date__isnull=True).exclude(
        Q(is_managed=True)
        | Q(dl_parameters__market_data__path="wbfdm.contrib.internal.dataloaders.market_data.MarketDataDataloader")
    )  # we exclude product and index managed to avoid circular import
    if instrument_ids:
        instruments = instruments.filter(id__in=instrument_ids)
    prices = list(instruments.get_instrument_prices_from_market_data(start, end))
    Instrument.bulk_save_instrument_prices(prices)
    investable_universe_updated.send(sender=Instrument, end_date=end)
    if with_background_tasks:
        daily_update_instrument_price_statistics.delay(from_date=start, to_date=end)
        update_instrument_metrics_as_task.delay()


@shared_task(queue=Queue.EXTENDED_BACKGROUND.value)
def update_instrument_metrics_as_task():
    instruments = Instrument.active_objects.filter(is_investable_universe=True)
    for instrument in tqdm(instruments, total=instruments.count()):
        instrument.update_last_valuation_date()
    instrument_metric_updated.send(sender=Instrument, basket=None, date=None, key=None)


@shared_task(queue=Queue.EXTENDED_BACKGROUND.value)
def synchronize_instruments_as_task():
    synchronize_instruments()


@shared_task(queue=Queue.BACKGROUND.value)
def synchronize_exchanges_as_task():
    synchronize_exchanges()


@shared_task(queue=Queue.EXTENDED_BACKGROUND.value)
def full_synchronization_as_task():
    # we get all instrument without name or where we would expect a parent and consider them for clean up.
    qs = Instrument.objects.filter(prices__isnull=True).filter(
        (Q(name="") & Q(name_repr="")) | (Q(source__in=["qa-ds2-security", "qa-ds2-quote"]) & Q(parent__isnull=True))
    )
    for instrument in qs:
        with suppress(ProtectedError):
            instrument.delete()
    mapping.cache_clear()  # we need to clear the mapping cache because we might have deleted parent instruments
    initialize_exchanges()
    initialize_instruments()
    with transaction.atomic():
        Instrument.objects.rebuild()  # rebuild MPTT tree


# Daily synchronization tasks.
# This tasks needs to be ran at maximum once a day in order to guarantee data consitency in
# case of change in change (e.g. reimport).
@shared_task(queue=Queue.BACKGROUND.value)
def daily_update_instrument_price_statistics(from_date: date = None, to_date: date = None):
    if not to_date:
        to_date = date.today()
    if not from_date:
        from_date = to_date - timedelta(days=3)
    # We query for the last 7 days unsynch instrument prices.
    prices = InstrumentPrice.objects.filter(date__gte=from_date, date__lte=to_date)
    objs = []
    for p in tqdm(prices.iterator(), total=prices.count()):
        p.compute_and_update_statistics()
        objs.append(p)
    InstrumentPrice.objects.bulk_update(
        objs,
        fields=["sharpe_ratio", "correlation", "beta", "annualized_daily_volatility", "volume_50d", "volume_200d"],
        batch_size=1000,
    )


@shared_task(queue=Queue.BACKGROUND.value)
def daily_refresh_instrument_list_constituent(val_date: date | None = None):
    if not val_date:
        val_date = date.today()

    for instrument_list in InstrumentList.objects.filter(constituent_loader_path__isnull=False):
        constituent = instrument_list.load_constituent(val_date)
        if len(constituent) > 0:
            instrument_list.refresh_constituent_from_list(val_date, list(constituent))



from .classifications import (
    ClassificationFilter,
    ClassificationTreeChartFilter,
    InstrumentClassificationThroughModelViewFilterSet,
)
from .financials_analysis import (
    FinancialAnalysisFilterSet,
    FinancialAnalysisValuationRatiosFilterSet,
    EarningsAnalysisFilterSet,
    GroupKeyFinancialsFilterSet,
)
from .instruments import (
    InstrumentFavoriteGroupFilterSet,
    InstrumentFilterSet,
    BaseClassifiedInstrumentFilterSet,
    MonthlyPerformancesInstrumentFilterSet,
)
from .instrument_prices import (
    InstrumentPriceFilterSet,
    InstrumentPriceSingleBenchmarkFilterSet,
    InstrumentPriceFrequencyFilter,
    InstrumentPriceFinancialStatisticsChartFilterSet,
    InstrumentPriceInstrumentFilterSet,
)
from .exchanges import ExchangeFilterSet
from .financials import (
    MarketDataChartFilterSet,
    FinancialRatioFilterSet,
    StatementFilter,
    StatementWithEstimateFilter,
)
from .instrument_list import InstrumentListThroughModelFilterSet



from datetime import timedelta

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from wbcore import viewsets
from wbcore.utils.date import get_date_interval_from_request

from wbfdm.figures.financials import FinancialsChartGenerator
from wbfdm.figures.financials.financial_analysis_charts import (
    FinancialAnalysisGenerator,
    PeriodChoices,
    VariableChoices,
)
from wbfdm.filters import (
    EarningsAnalysisFilterSet,
    FinancialAnalysisFilterSet,
    FinancialAnalysisValuationRatiosFilterSet,
)
from wbfdm.models import Instrument
from wbfdm.viewsets.configs import (
    CashFlowAnalysisInstrumentBarChartEndpointConfig,
    CashFlowAnalysisInstrumentBarChartTitle,
    CashFlowAnalysisInstrumentTableChartEndpointConfig,
    CashFlowAnalysisInstrumentTableChartTitle,
    EarningsInstrumentChartEndpointConfig,
    EarningsInstrumentChartTitle,
    FinancialAnalysisGeneratorTitleConfig,
    FinancialsGraphInstrumentChartEndpointConfig,
    FinancialsGraphInstrumentChartTitle,
    NetDebtAndEbitdaInstrumentChartEndpointConfig,
    NetDebtAndEbitdaInstrumentChartTitle,
    ProfitabilityRatiosInstrumentChartEndpointConfig,
    ProfitabilityRatiosInstrumentChartTitle,
    SummaryTableInstrumentChartEndpointConfig,
    SummaryTableInstrumentChartTitle,
)

from ..mixins import InstrumentMixin


class SummaryTableInstrumentChartViewSet(InstrumentMixin, viewsets.ChartViewSet):
    queryset = Instrument.objects.all()
    title_config_class = SummaryTableInstrumentChartTitle
    endpoint_config_class = SummaryTableInstrumentChartEndpointConfig
    filterset_class = FinancialAnalysisFilterSet

    def get_queryset(self):
        return Instrument.objects.filter(id=self.instrument.id)

    def get_plotly(self, queryset):
        instrument = queryset.first()
        instance_generator = FinancialsChartGenerator(instrument=instrument)
        fig = instance_generator.summary_table_chart()
        return fig


class FinancialsGraphInstrumentChartViewSet(SummaryTableInstrumentChartViewSet):
    title_config_class = FinancialsGraphInstrumentChartTitle
    endpoint_config_class = FinancialsGraphInstrumentChartEndpointConfig

    def get_plotly(self, queryset):
        instrument = queryset.first()
        instance_generator = FinancialsChartGenerator(instrument=instrument)
        fig = instance_generator.financials_chart()
        return fig


class ProfitabilityRatiosInstrumentChartViewSet(SummaryTableInstrumentChartViewSet):
    title_config_class = ProfitabilityRatiosInstrumentChartTitle
    endpoint_config_class = ProfitabilityRatiosInstrumentChartEndpointConfig

    def get_plotly(self, queryset):
        instrument = queryset.first()
        instance_generator = FinancialsChartGenerator(instrument=instrument)
        fig = instance_generator.profitability_ratios_chart()
        return fig


class CashFlowAnalysisInstrumentTableViewSet(SummaryTableInstrumentChartViewSet):
    title_config_class = CashFlowAnalysisInstrumentTableChartTitle
    endpoint_config_class = CashFlowAnalysisInstrumentTableChartEndpointConfig

    def get_plotly(self, queryset):
        instrument = queryset.first()
        instance_generator = FinancialsChartGenerator(instrument=instrument)
        fig = instance_generator.cash_flow_analysis_table_chart()
        return fig


class CashFlowAnalysisInstrumentBarChartViewSet(SummaryTableInstrumentChartViewSet):
    title_config_class = CashFlowAnalysisInstrumentBarChartTitle
    endpoint_config_class = CashFlowAnalysisInstrumentBarChartEndpointConfig

    def get_plotly(self, queryset):
        instrument = queryset.first()
        instance_generator = FinancialsChartGenerator(instrument=instrument)
        fig = instance_generator.cash_flow_analysis_bar_chart()
        return fig


class NetDebtAndEbitdaInstrumentChartViewSet(SummaryTableInstrumentChartViewSet):
    title_config_class = NetDebtAndEbitdaInstrumentChartTitle
    endpoint_config_class = NetDebtAndEbitdaInstrumentChartEndpointConfig

    def get_plotly(self, queryset):
        instrument = queryset.first()
        instance_generator = FinancialsChartGenerator(instrument=instrument)
        fig = instance_generator.net_debt_and_ebitda_chart()
        return fig


class ValuationRatiosChartView(InstrumentMixin, viewsets.ChartViewSet):
    queryset = Instrument.objects.all()
    title_config_class = FinancialAnalysisGeneratorTitleConfig
    filterset_class = FinancialAnalysisValuationRatiosFilterSet
    LIST_DOCUMENTATION = "wbfdm/markdown/documentation/financial_analysis_instrument_ratios.md"

    def get_plotly(self, queryset):  # noqa: C901
        # Set plotly as the default plotting lib
        pd.options.plotting.backend = "plotly"

        # GET data from fake filters
        date1, date2 = get_date_interval_from_request(self.request)
        if not date1 or not date2:
            return go.Figure()
        period = getattr(PeriodChoices, self.request.GET.get("period", "NTM"), PeriodChoices.NTM)
        output = self.request.GET.get("output", "CHART")
        vs_related = self.request.GET.get("vs_related", "false") == "true"
        clean_data = self.request.GET.get("clean_data", "true") == "true"
        ranges = self.request.GET.get("ranges", "false") == "true"
        range_type = self.request.GET.get("range_type", "MINMAX")
        range_period = int(self.request.GET.get("range_period", "120"))
        x_axis_var = getattr(VariableChoices, self.request.GET.get("x_axis_var", "EPSG"), VariableChoices.EPSG)
        x_axis, x_axis_title, x_axis_format = x_axis_var.lower(), x_axis_var.chart_label, x_axis_var.format
        y_axis_var = getattr(VariableChoices, self.request.GET.get("y_axis_var", "PE"), VariableChoices.PE)
        y_axis, y_axis_title, y_axis_format = y_axis_var.lower(), y_axis_var.chart_label, y_axis_var.format
        z_axis_var = getattr(VariableChoices, self.request.GET.get("z_axis_var", "MKTCAP"), VariableChoices.MKTCAP)
        z_axis, z_axis_title, z_axis_format = z_axis_var.lower(), z_axis_var.chart_label, z_axis_var.format
        median = self.request.GET.get("median", "true") == "true"

        # Fail early if filter combination doesn't make sense or the period is too large for the peer chart
        if output == "TSTABLE" and vs_related:
            fig = go.Figure()
            fig.update_layout(
                title_text="To provide the best user experience, you may not use <b>'versus related'</b> instruments in the <b>'time-series'</b> table view."
            )
            return fig

        if output == "CHART" and vs_related and ranges:
            fig = go.Figure()
            fig.update_layout(
                title_text="You may only <b>'draw ranges'</b> for a single instrument <b>('versus related' = NO)</b>."
            )
            return fig

        if output == "CHART" and vs_related and (date2 - date1).days > 900:
            fig = go.Figure()
            fig.update_layout(
                title_text="To provide the best user experience, it is possible to display <b>up to 2.5 calendar years</b> in the chart view with <b>'versus related = YES'</b>."
            )
            return fig

        if output == "CHART" and ranges and range_type == "ROLLING" and range_period < 1:
            fig = go.Figure()
            fig.update_layout(title_text="<b>Rolling period</b> cannot be lower than 1.")
            return fig

        # Get all ratios
        generator = FinancialAnalysisGenerator([self.instrument], date1, date2, vs_related)
        ratios = generator.get_common_valuation_ratios_df(period, clean_data)

        # If ratios are empty, return early
        if ratios.empty:
            fig = go.Figure()
            fig.update_layout(title_text="No data available.")
            return fig

        # Label dates
        ratios["datetxt"] = pd.to_datetime(ratios["date"]).dt.strftime("%Y-%m-%d")

        if output == FinancialAnalysisValuationRatiosFilterSet.OutputChoices.TSTABLE:
            fig = go.Figure(
                data=[
                    go.Table(
                        header=dict(
                            values=[
                                "Date",
                                "Price-to-Earnings PE " + period,
                                "Price-to-Earnings-to-Growth PEG " + period,
                                "Price-to-Sales PS " + period,
                                "EV-to-EBITDA " + period,
                                "Price-to-FreeCashflow P/FCF " + period,
                            ]
                        ),
                        cells=dict(
                            values=[
                                ratios["datetxt"],
                                ratios["pe"].astype(float).round(1).replace([np.nan], None).replace([None], ""),
                                ratios["peg"].astype(float).round(2).replace([np.nan], None).replace([None], ""),
                                ratios["ps"].astype(float).round(1).replace([np.nan], None).replace([None], ""),
                                ratios["evebitda"].astype(float).round(1).replace([np.nan], None).replace([None], ""),
                                ratios["pfcf"].astype(float).round(1).replace([np.nan], None).replace([None], ""),
                            ],
                        ),
                    ),
                ]
            )
            return fig

        elif output == FinancialAnalysisValuationRatiosFilterSet.OutputChoices.TABLE:
            ratios = ratios.groupby("instrument").tail(1)
            ratios = ratios.reset_index(drop=True)

            # If table and additional instruments, add median and mean rows
            if vs_related:
                ratios.fillna(value=np.nan, inplace=True)
                ratios_median = ratios.median(numeric_only=True, axis=0)
                ratios.loc["Mean"] = ratios.mean(numeric_only=True, axis=0)
                ratios.loc[ratios.index[-1], "instrument_title"] = "Mean"
                ratios.loc["Median"] = ratios_median
                ratios.loc[ratios.index[-1], "instrument_title"] = "Median"
                ratios.rename(index={"Mean": ratios["date"][0]}, inplace=True)
                ratios.rename(index={"Median": ratios["date"][0]}, inplace=True)
                ratios.replace(np.nan, None, inplace=True)

            # Round all variables to 1 decimal but peg
            columns = ratios.columns.difference(["instrument", "instrument_title", "peg", "date", "datetxt"])
            for col in columns:
                ratios[col] = ratios[col].astype(float).round(1).replace([np.nan], None).replace([None], "")
            ratios["peg"] = ratios["peg"].astype(float).round(2).replace([np.nan], None).replace([None], "")

            # last two lines in bold
            if vs_related:
                for i in [0, -2, -1]:
                    row = ratios.iloc[i, :].values
                    bold_row = ["<b>" + str(entry) + "</b>" for entry in row]
                    ratios.iloc[i, :] = bold_row

            return go.Figure(
                data=[
                    go.Table(
                        header=dict(
                            values=[
                                period + " on " + str(ratios["datetxt"].head(1).item()),
                                "Price-to-Earnings PE",
                                "Price-to-Earnings-to-Growth PEG",
                                "Price-to-Sales PS",
                                "EV-to-EBITDA",
                                "Price-to-FreeCashFlow P/FCF",
                            ]
                        ),
                        cells=dict(
                            values=[
                                ratios["instrument_title"],
                                ratios["pe"],
                                ratios["peg"],
                                ratios["ps"],
                                ratios["evebitda"],
                                ratios["pfcf"],
                            ],
                        ),
                    )
                ]
            )

        else:
            # Calculate custom axis ranges for better visual representation
            def axis_range(series, type="lower"):
                if type == "lower":
                    a = series.min()
                    if pd.isna(a):
                        return 0
                    if a > 0:
                        _range = 0.75 * float(str(a).replace("nan", "0"))
                    else:
                        _range = 1.25 * float(str(a).replace("nan", "-1"))
                else:
                    a = series.max()
                    if pd.isna(a):
                        return 1
                    if a > 0:
                        _range = 1.25 * float(str(a).replace("nan", "30"))
                    else:
                        _range = 0.75 * float(str(a).replace("nan", "0"))
                return _range

            if vs_related:
                # Get biweekly dates to run the loop for plotly sliders
                def daterange(start_date, end_date):
                    for n in range(1 + int((end_date - start_date).days / 14)):
                        if n == len(range(int((end_date - start_date).days / 14))):
                            yield end_date
                        yield start_date + timedelta(14 * n)

                fig = go.Figure()
                dates = []
                for single_date in daterange(date1, date2):
                    ratio_trace = ratios[
                        (ratios["datetxt"] == single_date.strftime("%Y-%m-%d")) & (ratios[z_axis] > 0)
                    ]
                    if (
                        not ratio_trace.empty
                        and not all(v is None for v in ratio_trace[x_axis])
                        and not all(v is None for v in ratio_trace[y_axis])
                    ):
                        dates.append(single_date.strftime("%Y-%m-%d"))
                        ratio_main = ratio_trace.head(1)
                        ratio_other = ratio_trace.iloc[1:]
                        fig.add_trace(
                            go.Scatter(
                                visible=False,
                                x=ratio_main[x_axis],
                                y=ratio_main[y_axis],
                                customdata=np.stack((ratio_main["instrument_title"], ratio_main[x_axis]), axis=-1),
                                mode="markers",
                                marker=dict(
                                    color="green",
                                    opacity=0.5,
                                    size=ratio_main[z_axis].astype(float),
                                    sizeref=(2.0 * np.nanmax(ratios[z_axis].fillna(0))) / (20.0**2),
                                ),
                                hovertemplate="<br>".join(
                                    [
                                        s.replace(" ", "&nbsp;")
                                        for s in [
                                            "<b>%{customdata[0]}</b>",
                                            "X:          <b>" + x_axis_title + ":</b>    %{x:" + str(x_axis_format),
                                            "Y:          <b>" + y_axis_title + ":</b>    %{y:" + str(y_axis_format),
                                            "Bubble: <b>"
                                            + z_axis_title
                                            + ":</b>    %{marker.size:"
                                            + str(z_axis_format)
                                            + "<extra></extra>",
                                        ]
                                    ]
                                ),
                            )
                        )
                        colors = px.colors.qualitative.Pastel
                        fig.add_trace(
                            go.Scatter(
                                visible=False,
                                name="Ratios: " + str(ratio_trace["datetxt"]),
                                x=ratio_other[x_axis],
                                y=ratio_other[y_axis],
                                customdata=np.stack((ratio_other["instrument_title"], ratio_other[x_axis]), axis=-1),
                                mode="markers",
                                marker=dict(
                                    color=colors[0 : len(ratio_other["instrument_title"])],
                                    opacity=0.5,
                                    size=ratio_other[z_axis].astype(float),
                                    sizeref=(2.0 * np.nanmax(ratios[z_axis].fillna(0))) / (20.0**2),
                                ),
                                hovertemplate="<br>".join(
                                    [
                                        s.replace(" ", "&nbsp;")
                                        for s in [
                                            "<b>%{customdata[0]}</b>",
                                            "X:          <b>" + x_axis_title + ":</b>    %{x:" + str(x_axis_format),
                                            "Y:          <b>" + y_axis_title + ":</b>    %{y:" + str(y_axis_format),
                                            "Bubble: <b>"
                                            + z_axis_title
                                            + ":</b>    %{marker.size:"
                                            + str(z_axis_format)
                                            + "<extra></extra>",
                                        ]
                                    ]
                                ),
                            )
                        )
                        fig.update_layout(title_text="Relative Valuation <b>" + period + "</b>", showlegend=False)
                        fig.update_xaxes(
                            title_text=x_axis_title + " <b>" + period + "</b>",
                            hoverformat=x_axis_format,
                            tickformat=x_axis_format.split("}")[0],
                            range=[axis_range(ratios[x_axis], type="lower"), axis_range(ratios[x_axis], type="upper")],
                        )
                        fig.update_yaxes(
                            title_text=y_axis_title + " <b>" + period + "</b>",
                            hoverformat=y_axis_title,
                            tickformat=y_axis_format.split("}")[0],
                            range=[axis_range(ratios[y_axis], type="lower"), axis_range(ratios[y_axis], type="upper")],
                        )
                        fig.add_hline(y=0, line_width=0.15, line_color="grey")
                        fig.add_vline(x=0, line_width=0.15, line_color="grey")
                        if median:
                            fig.add_hline(y=ratios[y_axis].median(), line_width=1, line_dash="dash", line_color="grey")
                            fig.add_vline(x=ratios[x_axis].median(), line_width=1, line_dash="dash", line_color="grey")

                if len(dates) == 0:
                    fig.update_layout(title_text="No data available.")
                    return fig
                else:
                    steps = []
                    for i in range(len(dates)):
                        step = dict(
                            method="update",
                            args=[
                                {"visible": [False] * len(dates) * 2},
                                {"name": "Relative Valuation <b>" + period + "</b>. Chosen date: " + str(dates[i])},
                            ],
                        )
                        step["args"][0]["visible"][i * 2] = True
                        step["args"][0]["visible"][(i * 2) + 1] = True
                        steps.append(step)

                    sliders = [
                        dict(active=len(dates) - 1, currentvalue={"prefix": "Date: "}, pad={"t": 50}, steps=steps)
                    ]

                    fig.data[-1].visible = True
                    fig.data[-2].visible = True

                    fig.update_layout(sliders=sliders)

                    for i in range(len(dates)):
                        fig["layout"]["sliders"][0]["steps"][i]["label"] = str(dates[i])

                    return fig

            if not ranges:
                fig = make_subplots(specs=[[{"secondary_y": True}]])

                for plot in [
                    ("pe", "#ff6361", VariableChoices.PE.chart_label, VariableChoices.PE, False),
                    ("peg", "#ffa600", VariableChoices.PEG.chart_label, VariableChoices.PEG, True),
                    ("ps", "#58508d", VariableChoices.PS.chart_label, VariableChoices.PS, False),
                    ("evebitda", "#003f5c", VariableChoices.EVEBITDA.chart_label, VariableChoices.EVEBITDA, False),
                    ("pfcf", "#bc5090", VariableChoices.PFCF.chart_label, VariableChoices.PFCF, False),
                ]:
                    fig.add_trace(
                        go.Scatter(
                            x=ratios["date"],
                            y=ratios[plot[0]],
                            line=dict(color=plot[1], width=2),
                            name=plot[2] + " " + period,
                            hovertemplate="%{x}: <b>" + plot[3] + " " + period + "</b> %{y} <extra></extra>",
                        ),
                        secondary_y=plot[4],
                    )

                fig.update_layout(title_text="Relative Valuation " + period, yaxis_ticksuffix="x")
                fig.update_yaxes(title_text="Ratios " + period, hoverformat=".2f", secondary_y=False)
                if abs(axis_range(ratios["peg"], type="upper")) >= 1:
                    ticks = ".1f"
                else:
                    ticks = ".2f"
                fig.update_yaxes(
                    title_text="PEG " + period,
                    hoverformat=".2f",
                    tickformat=ticks,
                    title_font_color="#ffa600",
                    secondary_y=True,
                )
                fig.update_yaxes(
                    rangemode="tozero", scaleanchor="y2", scaleratio=0.1, constraintoward="bottom", secondary_y=False
                )
                fig.update_yaxes(
                    rangemode="tozero", scaleanchor="y", scaleratio=1, constraintoward="bottom", secondary_y=True
                )
                fig.add_hline(y=0, line_width=3, line_dash="dash", line_color="grey")
                fig["layout"]["yaxis2"]["showgrid"] = False
                return fig

            fig = make_subplots(rows=5, cols=1, vertical_spacing=0.01, shared_xaxes=True)
            x_range = [ratios["date"].iloc[0], ratios["date"].iloc[-1]]

            for index, plot in enumerate(
                [
                    ("pe", "#ff6361"),
                    ("peg", "#ffa600"),
                    ("ps", "#58508d"),
                    ("evebitda", "#003f5c"),
                    ("pfcf", "#bc5090"),
                ],
                start=1,
            ):
                fig.add_trace(
                    go.Scatter(x=ratios["date"], y=ratios[plot[0]], line=dict(color=plot[1], width=2)),
                    row=index,
                    col=1,
                )
                if range_type == FinancialAnalysisValuationRatiosFilterSet.RangeChoices.MINMAX:
                    if not ratios[plot[0]].isnull().all():
                        fig.add_trace(
                            go.Scatter(
                                x=x_range,
                                y=[ratios[plot[0]].max(), ratios[plot[0]].max()],
                                mode="lines",
                                line_width=1,
                                line_dash="dash",
                                line_color=plot[1],
                            ),
                            row=index,
                            col=1,
                        )
                        fig.add_trace(
                            go.Scatter(
                                x=x_range,
                                y=[ratios[plot[0]].min(), ratios[plot[0]].min()],
                                mode="lines",
                                line_width=1,
                                line_dash="dash",
                                line_color=plot[1],
                            ),
                            row=index,
                            col=1,
                        )
                else:
                    fig.add_trace(
                        go.Scatter(
                            x=ratios["date"],
                            y=ratios[plot[0]].rolling(range_period, min_periods=1).max().fillna(""),
                            mode="lines",
                            line_width=0.75,
                            line_dash="dash",
                            line_color=plot[1],
                        ),
                        row=index,
                        col=1,
                    )
                    fig.add_trace(
                        go.Scatter(
                            x=ratios["date"],
                            y=ratios[plot[0]].rolling(range_period, min_periods=1).min().fillna(""),
                            mode="lines",
                            line_width=0.75,
                            line_dash="dash",
                            line_color=plot[1],
                        ),
                        row=index,
                        col=1,
                    )

            fig.update_layout(
                yaxis=dict(title_text="P/E", ticksuffix="x", hoverformat=".1f"),
                yaxis2=dict(title_text="PEG", tickformat=".1f", hoverformat=".1f"),
                yaxis3=dict(title_text="P/S", ticksuffix="x", hoverformat=".1f"),
                yaxis4=dict(title_text="EV/EBITDA", ticksuffix="x", hoverformat=".1f"),
                yaxis5=dict(title_text="P/FCF", ticksuffix="x", hoverformat=".1f"),
            )

            fig.update_traces(hovertemplate="%{x}: %{y} <extra></extra>")

            if range_type == FinancialAnalysisValuationRatiosFilterSet.RangeChoices.MINMAX:
                fig.update_layout(
                    title_text=period + " Valuation ratios with Min/Max range", title_x=0, showlegend=False
                )
            else:
                fig.update_layout(
                    title_text=period + " Valuation ratios with " + str(range_period) + " days rolling min-max range",
                    title_x=0,
                    showlegend=False,
                )

            return fig


class EarningsInstrumentChartViewSet(InstrumentMixin, viewsets.ChartViewSet):
    title_config_class = EarningsInstrumentChartTitle
    endpoint_config_class = EarningsInstrumentChartEndpointConfig
    queryset = Instrument.objects.all()
    filterset_class = EarningsAnalysisFilterSet
    LIST_DOCUMENTATION = "wbfdm/markdown/documentation/earnings_instrument.md"

    def get_plotly(self, queryset):
        # GET data from fake filters
        date1, date2 = get_date_interval_from_request(self.request)
        if not date1 or not date2:
            return go.Figure()
        period = getattr(PeriodChoices, self.request.GET.get("period", "NTM"), PeriodChoices.NTM)
        analysis = self.request.GET.get("output", "EPS")
        vs_related = self.request.GET.get("vs_related", "false") == "true"

        # Get all earnings data
        generator = FinancialAnalysisGenerator([self.instrument], date1, date2, vs_related)
        earnings = generator.get_earnings_df(period, clean_data=True)

        # If earnings are empty, return early
        if earnings.empty:
            fig = go.Figure()
            fig.update_layout(title_text="No data available.")
            return fig

        if period == PeriodChoices.TTM:
            hover_helper = " EPS "
            title_helper = "Reported Earnings "
        else:
            hover_helper = " Consensus EPS "
            title_helper = "Consensus Earnings Estimates "

        fig = go.Figure()
        earnings = earnings.set_index(["instrument", "instrument_title", "date"]).unstack(level=1)
        earnings = earnings.reset_index().drop("instrument", axis=1).replace(np.nan, None).set_index("date")
        if analysis == "EPS":
            for col in earnings.columns:
                fig.add_trace(
                    go.Scatter(
                        x=earnings.index,
                        y=earnings[col],
                        name=col[1],
                        hovertemplate="%{x}: <b>" + col[1] + hover_helper + period + "</b> %{y} <extra></extra>",
                    ),
                )

            fig.update_layout(title_text=title_helper + period, yaxis_tickprefix="$")
            fig.update_yaxes(title_text="EPS ($) " + period, hoverformat=".2f", tickformat=".2f")

        return fig



import pytest
from django.forms.models import model_to_dict
from faker import Faker

from wbfdm.import_export.handlers.option import (
    OptionAggregateImportHandler,
    OptionImportHandler,
)
from wbfdm.models import Option, OptionAggregate

fake = Faker()


@pytest.mark.django_db
class TestOptionAggregateModel:
    def test_option_creation(self, option_aggregate):
        assert option_aggregate.id is not None

    def test_import_option(self, import_source, instrument, option_aggregate_factory):
        def serialize(option_aggregate):
            data = model_to_dict(option_aggregate)
            data["date"] = option_aggregate.date.strftime("%Y-%m-%d")
            data["instrument"] = {"id": instrument.id}
            del data["id"]
            del data["import_source"]
            return data

        option_aggregate = option_aggregate_factory.build()
        data = {"data": [serialize(option_aggregate)]}
        handler = OptionAggregateImportHandler(import_source)

        # Import non existing data
        handler.process(data)
        assert OptionAggregate.objects.count() == 1

        # Import already existing data
        # import_source.data['data'][0]['shares'] *= 2

        handler.process(data)
        assert OptionAggregate.objects.count() == 1


@pytest.mark.django_db
class TestOptionModel:
    def test_option_creation(self, option):
        assert option.id is not None

    def test_import_option(self, import_source, instrument, option_factory):
        def serialize(option):
            data = model_to_dict(option)
            data["date"] = option.date.strftime("%Y-%m-%d")
            data["expiration_date"] = option.expiration_date.strftime("%Y-%m-%d")
            data["instrument"] = {"id": instrument.id}
            del data["id"]
            del data["import_source"]
            return data

        option = option_factory.build()
        data = {"data": [serialize(option)]}
        handler = OptionImportHandler(import_source)

        # Import non existing data
        handler.process(data)
        assert Option.objects.count() == 1

        # Import already existing data
        # import_source.data['data'][0]['shares'] *= 2

        handler.process(data)
        assert Option.objects.count() == 1



from wbcore import viewsets
from wbcore.contrib.permission.permissions import InternalUserPermissionMixin

from wbfdm.filters import ExchangeFilterSet
from wbfdm.models import Exchange
from wbfdm.serializers import ExchangeModelSerializer, ExchangeRepresentationSerializer

from .configs import (
    ExchangeButtonConfig,
    ExchangeDisplayConfig,
    ExchangeEndpointConfig,
    ExchangeTitleConfig,
)


class ExchangeRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    queryset = Exchange.objects.all()
    serializer_class = ExchangeRepresentationSerializer

    search_fields = (
        "name",
        "mic_code",
        "operating_mic_code",
        "bbg_exchange_codes",
        "bbg_composite_primary",
        "bbg_composite",
        "refinitiv_mnemonic",
        "refinitiv_identifier_code",
        "city__name",
        "website",
        "comments",
    )
    ordering_fields = search_fields
    ordering = ["name"]


class ExchangeModelViewSet(InternalUserPermissionMixin, viewsets.ModelViewSet):
    filterset_class = ExchangeFilterSet
    serializer_class = ExchangeModelSerializer
    queryset = Exchange.objects.select_related(
        "country",
        "city",
    )

    search_fields = (
        "name",
        "mic_code",
        "operating_mic_code",
        "bbg_exchange_codes",
        "bbg_composite_primary",
        "refinitiv_mnemonic",
        "bbg_composite",
        "refinitiv_identifier_code",
        "city__name",
        "website",
        "comments",
    )
    ordering_fields = search_fields
    ordering = ["name"]

    display_config_class = ExchangeDisplayConfig
    button_config_class = ExchangeButtonConfig
    title_config_class = ExchangeTitleConfig
    endpoint_config_class = ExchangeEndpointConfig



from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.content_type.serializers import ContentTypeRepresentationSerializer

from wbfdm.contrib.metric.models import InstrumentMetric
from wbfdm.serializers.instruments import InstrumentRepresentationSerializer


class InstrumentMetricRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = InstrumentMetric
        fields = ("id", "key", "date", "basket_repr")


class InstrumentMetricModelSerializer(wb_serializers.ModelSerializer):
    _basket_content_type = ContentTypeRepresentationSerializer(source="basket_content_type")
    _instrument = InstrumentRepresentationSerializer(source="instrument")
    _parent_metric = InstrumentMetricRepresentationSerializer(source="parent_metric")
    _group_key = wb_serializers.CharField(read_only=True)

    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        return {
            "children_metrics": f"{reverse('metric:instrumentmetric-list', args=[], request=request)}?parent_metric={instance.id}"
        }

    class Meta:
        model = InstrumentMetric
        fields = (
            "id",
            "basket_content_type",
            "_basket_content_type",
            "basket_repr",
            "basket_id",
            "instrument",
            "_instrument",
            "date",
            "key",
            "metrics",
            "parent_metric",
            "_parent_metric",
            "_group_key",
            "_additional_resources",
        )
        read_only_fields = fields



from datetime import date
from decimal import Decimal

import pandas as pd
import pytest

from wbfdm.models import Instrument


@pytest.mark.django_db
class TestInstrumentQueryset:
    def test_get_returns(self, instrument_factory, instrument_price_factory):
        v1 = date(2024, 12, 31)
        v2 = date(2025, 1, 1)
        v3 = date(2025, 1, 2)
        v4 = date(2025, 1, 3)

        i1 = instrument_factory.create()
        i2 = instrument_factory.create()

        i11 = instrument_price_factory.create(date=v1, instrument=i1)
        i12 = instrument_price_factory.create(date=v2, instrument=i1)
        i14 = instrument_price_factory.create(date=v4, instrument=i1)
        i21 = instrument_price_factory.create(date=v1, instrument=i2)
        i11.refresh_from_db()
        i12.refresh_from_db()
        i14.refresh_from_db()
        prices, returns = Instrument.objects.filter(id__in=[i1.id, i2.id]).get_returns_df(from_date=v1, to_date=v4)

        expected_returns = pd.DataFrame(
            [[i12.net_value / i11.net_value - 1, 0.0], [0.0, 0.0], [i14.net_value / i12.net_value - 1, 0.0]],
            index=[v2, v3, v4],
            columns=[i1.id, i2.id],
            dtype="float64",
        )
        expected_returns.index = pd.to_datetime(expected_returns.index)
        pd.testing.assert_frame_equal(returns, expected_returns, check_names=False, check_freq=False, atol=1e-6)
        assert prices[v1][i1.id] == float(i11.net_value)
        assert prices[v2][i1.id] == float(i12.net_value)
        assert prices[v3][i1.id] == float(i12.net_value)
        assert prices[v4][i1.id] == float(i14.net_value)

        prices, returns = Instrument.objects.filter(id__in=[i1.id, i2.id]).get_returns_df(
            from_date=v1, to_date=v4, ffill_period=2
        )
        # test that the returned price are ffill up to the specified ffill_periord
        assert prices[v1][i2.id] == float(i21.net_value)
        assert prices[v2][i2.id] == float(i21.net_value)
        assert prices[v3][i2.id] == float(i21.net_value)
        assert prices[v4][i2.id] is None

        # Check that a delisted instrument won't be ffilled after its delisted date
        i2.delisted_date = v2
        i2.save()
        prices, returns = Instrument.objects.filter(id__in=[i1.id, i2.id]).get_returns_df(
            from_date=v1, to_date=v4, ffill_period=10
        )
        assert prices[v1][i2.id] == float(i21.net_value)
        assert prices[v2][i2.id] == float(i21.net_value)
        assert prices[v3][i2.id] is None
        assert prices[v4][i2.id] is None

    def test_get_returns_fix_forex_on_holiday(
        self, instrument, instrument_price_factory, currency_fx_rates_factory, currency_factory
    ):
        v1 = date(2024, 12, 31)
        v2 = date(2025, 1, 1)
        v3 = date(2025, 1, 2)

        target_currency = currency_factory.create()
        fx_target1 = currency_fx_rates_factory.create(currency=target_currency, date=v1)
        fx_target2 = currency_fx_rates_factory.create(currency=target_currency, date=v2)  # noqa
        fx_target3 = currency_fx_rates_factory.create(currency=target_currency, date=v3)

        fx1 = currency_fx_rates_factory.create(currency=instrument.currency, date=v1)
        fx2 = currency_fx_rates_factory.create(currency=instrument.currency, date=v2)  # noqa
        fx3 = currency_fx_rates_factory.create(currency=instrument.currency, date=v3)

        i1 = instrument_price_factory.create(net_value=Decimal("100"), date=v1, instrument=instrument)
        i2 = instrument_price_factory.create(net_value=Decimal("100"), date=v2, instrument=instrument, calculated=True)
        i3 = instrument_price_factory.create(net_value=Decimal("200"), date=v3, instrument=instrument)

        prices, returns = Instrument.objects.filter(id__in=[instrument.id]).get_returns_df(
            from_date=v1, to_date=v3, to_currency=target_currency
        )
        returns.index = pd.to_datetime(returns.index)
        assert prices[v1][instrument.id] == float(i1.net_value)
        assert prices[v2][instrument.id] == float(i2.net_value)
        assert prices[v3][instrument.id] == float(i3.net_value)

        assert returns.loc[pd.Timestamp(v2), instrument.id] == pytest.approx(
            float(
                (i2.net_value * fx_target1.value / fx1.value) / (i1.net_value * fx_target1.value / fx1.value)
                - Decimal("1")
            ),
            abs=10e-8,
        )  # as v2 as a calculated price, the forex won't apply to it
        assert returns.loc[pd.Timestamp(v3), instrument.id] == pytest.approx(
            float(
                (i3.net_value * fx_target3.value / fx3.value) / (i2.net_value * fx_target1.value / fx1.value)
                - Decimal("1")
            ),
            abs=10e-8,
        )



from wbcore.menus import ItemPermission, MenuItem

INSTRUMENT_MENUITEM = MenuItem(
    label="Instrument",
    endpoint="wbfdm:instrument-list",
    endpoint_get_parameters={"parent__isnull": True},
    permission=ItemPermission(method=lambda request: request.user.is_internal, permissions=["wbfdm.view_instrument"]),
)

INVESTABLE_UNIVERSE_MENUITEM = MenuItem(
    label="Investable Universe",
    endpoint="wbfdm:instrument-list",
    endpoint_get_parameters={"is_investable_universe": True},
    permission=ItemPermission(method=lambda request: request.user.is_internal, permissions=["wbfdm.view_instrument"]),
)



from django.contrib.postgres.search import SearchQuery, SearchRank
from django.db.models import F, Q
from django.db.models.expressions import Value
from django.db.models.functions import Coalesce


class InstrumentSearchFilter:
    def get_min_search_rank(self, view) -> float:
        return getattr(view, "SEARCH_MIN_RANK", 0.3)

    def filter_queryset(self, request, queryset, view):
        if search := request.GET.get("search", None):
            min_search_rank = self.get_min_search_rank(view)
            query = SearchQuery(search, search_type="phrase")
            return (
                queryset.annotate(search_rank=Coalesce(SearchRank(F("search_vector"), query), Value(-1.0)))
                .filter(
                    (Q(search_vector=query) & Q(search_rank__gte=min_search_rank))
                    | Q(name__icontains=search)
                    | Q(name_repr__icontains=search)
                    | Q(isin__icontains=search)
                )
                .order_by("-search_rank")
            ).distinct()
        elif "parent" in request.GET:
            return queryset.order_by("-is_primary", "computed_str")
        return queryset



from contextlib import suppress
from datetime import UTC, datetime
from typing import Generator

import requests
from django.core.cache import cache as cache_layer
from django.utils import timezone
from django.utils.functional import cached_property
from jwt import decode as jwt_decode
from jwt.exceptions import DecodeError
from requests.exceptions import ConnectionError


def is_expired(token: str) -> bool:
    with suppress(DecodeError, KeyError, ValueError):
        expiry_ts = int(jwt_decode(token, options={"verify_signature": False})["exp"])
        expiry_datetime = datetime.fromtimestamp(expiry_ts, UTC)
        return expiry_datetime < timezone.now()
    return True


class MSCIClient:
    AUTH_SERVER_ULR: str = "https://accounts.msci.com/oauth/token"

    def __init__(self, client_id: str, client_secret: str):
        self.client_id = client_id
        self.client_secret = client_secret

    @cached_property
    def oauth_token(self) -> str:
        token = cache_layer.get("msci_oauth_token")
        if not token or is_expired(token):
            resp = requests.post(
                self.AUTH_SERVER_ULR,
                json={
                    "client_id": self.client_id,
                    "client_secret": self.client_secret,
                    "grant_type": "client_credentials",
                    "audience": "https://esg/data",
                },
                timeout=10,
            )
            if resp.status_code == requests.codes.ok:
                with suppress(KeyError, requests.exceptions.JSONDecodeError):
                    resp_json = resp.json()
                    token = resp_json["access_token"]
                    expires_in = resp_json["expires_in"]
                    cache_layer.set("msci_oauth_token", token, expires_in)
        if token:
            return token
        raise ConnectionError()

    def esg(self, identifiers: list[str], factors: list[str]) -> Generator[dict[str, str], None, None]:
        with suppress(ConnectionError):
            response = requests.post(
                url="https://api2.msci.com/esg/data/v2.0/issuers",
                json={
                    "issuer_identifier_list": identifiers,
                    "factor_name_list": factors,
                },
                headers={"AUTHORIZATION": f"Bearer {self.oauth_token}"},
                timeout=10,
            )
            if response.ok:
                for row in response.json().get("result", {}).get("issuers", []):
                    yield row

    def controversies(self, identifiers: list[str], factors: list[str]) -> Generator[dict[str, str], None, None]:
        next_url = "https://api2.msci.com/esg/data/v2.0/issuers"
        offset = 0
        limit = 100
        while next_url:
            with suppress(ConnectionError):
                response = requests.post(
                    url=next_url,
                    json={
                        "issuer_identifier_list": identifiers,
                        "factor_name_list": factors,
                        "limit": limit,
                        "offset": offset,
                    },
                    headers={"AUTHORIZATION": f"Bearer {self.oauth_token}"},
                    timeout=10,
                )

                if not response.ok:
                    next_url = None
                else:
                    json_res = response.json()
                    try:
                        next_url = json_res["paging"]["links"]["next"]
                    except KeyError:
                        next_url = None
                    offset += limit
                    for row in json_res.get("result", {}).get("issuers", []):
                        yield row



import pandas as pd
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf

from wbfdm.enums import ESG
from wbfdm.models.instruments import Instrument
from wbfdm.serializers import InstrumentControversySerializer
from wbfdm.viewsets.configs import (
    InstrumentESGControversiesEndpointViewConfig,
    InstrumentESGControversiesTitleViewConfig,
    InstrumentESGControversyDisplayViewConfig,
    InstrumentESGPAIDisplayViewConfig,
    InstrumentESGPAIEndpointViewConfig,
    InstrumentESGPAITitleViewConfig,
)

from .mixins import InstrumentMixin


class InstrumentESGControversiesViewSet(InstrumentMixin, viewsets.ViewSet):
    IDENTIFIER = "wbfdm:instrument-controversy"
    display_config_class = InstrumentESGControversyDisplayViewConfig
    title_config_class = InstrumentESGControversiesTitleViewConfig
    endpoint_config_class = InstrumentESGControversiesEndpointViewConfig
    serializer_class = InstrumentControversySerializer
    queryset = Instrument.objects.none()
    permission_classes = []

    def list(self, request, instrument_id):
        queryset = self.get_queryset()
        serializer = self.get_serializer_class()
        serializer = serializer(queryset, many=True)
        return Response({"results": serializer.data})

    def get_queryset(self):
        return sorted(
            Instrument.objects.filter(id=self.instrument.id).dl.esg_controversies(),
            key=lambda x: (x["review"] is None, x["review"]),
            reverse=True,
        )


class InstrumentESGPAIViewSet(InstrumentMixin, ExportPandasAPIViewSet):
    queryset = Instrument.objects.none()
    display_config_class = InstrumentESGPAIDisplayViewConfig
    endpoint_config_class = InstrumentESGPAIEndpointViewConfig
    title_config_class = InstrumentESGPAITitleViewConfig

    def get_queryset(self):
        return Instrument.objects.filter(id=self.instrument.id)

    def get_pandas_fields(self, request):
        return pf.PandasFields(
            fields=[
                pf.PKField(key="index", label="ID"),
                pf.IntegerField(key="section", label="Section"),
                pf.TextField(key="asi", label="Adverse sustainability indicator"),
                pf.TextField(key="metric", label="Metric"),
                pf.TextField(key="factor", label="Factor"),
                pf.FloatField(key="value", label="Value"),
            ]
        )

    def get_dataframe(self, request, queryset, **kwargs):
        df = pd.DataFrame(queryset.dl.esg(values=list(ESG))).reset_index()
        if not df.empty:
            esg_mapping = ESG.mapping()
            df[["section", "asi", "metric", "factor"]] = pd.DataFrame(
                df.factor_code.map(esg_mapping).tolist(), index=df.index
            )
        return df



import math
import re
from datetime import date, datetime
from typing import Generator, List, Optional

import DatastreamPy as dsweb  # noqa
import numpy as np
import pandas as pd
import pytz
from django.core.cache import cache
from django.utils import timezone
from wbcore.contrib.currency.models import Currency, CurrencyFXRates


class CachedTokenDataClient(dsweb.DataClient):
    def _get_token(self, isProxy=False):  # noqa
        if (token := cache.get("dsws_token")) and (token_expiry := cache.get("dsws_token_expiry")):
            self.token = token
            self.tokenExpiry = timezone.make_aware(datetime.fromtimestamp(token_expiry), timezone=pytz.UTC)
        else:
            super()._get_token()
            expiry_elapse = (self.tokenExpiry - timezone.now()).seconds
            cache.set("dsws_token", self.token, timeout=expiry_elapse)
            cache.set("dsws_token_expiry", datetime.timestamp(self.tokenExpiry), timeout=expiry_elapse)


class Client:
    MAXIMUM_ITEMS_PER_BUNDLE: int = 500
    MAXIMUM_REQUESTS_PER_BUNDLE: int = 20
    MAXIMUM_INSTRUMENTS_PER_REQUEST: int = 50
    MAXIMUM_DATATYPES_PER_REQUEST: int = 50
    MAXIMUM_ITEMS_PER_REQUEST: int = 100
    PRICE_ERROR_MARGIN: float = 0.10
    IBUNIT_DEFAULT_UNIT: float = 1e6

    def __init__(self, username: Optional[str] = None, password: Optional[str] = None):
        self.connection = CachedTokenDataClient(None, username, password)
        super().__init__()

    @classmethod
    def get_chunked_list(cls, identifiers: list[str], fields_number: int) -> list[list[str]]:
        instruments_number = len(identifiers)
        max_chunks_number = instruments_number * fields_number / cls.MAXIMUM_ITEMS_PER_BUNDLE
        chunk_size = int(instruments_number / max_chunks_number) - 1
        res = []
        for x in range(0, len(identifiers), chunk_size):
            res.append(identifiers[x : x + chunk_size])
        return res

    @classmethod
    def _breakdown_requests(
        cls, tickers: List[str], symbols: List[str]
    ) -> Generator[tuple[list[str], list[str]], None, None]:
        """
        Helper function to generate tuple of ticker and symbols that respect the API data usage
        Args:
            tickers: List of tickers to fetch
            symbols: Corresponding list of symbol to fetch

        Returns:
            Yield tuple of list of string
        """
        if len(tickers) * len(symbols) > cls.MAXIMUM_ITEMS_PER_BUNDLE:
            raise ValueError(f"The maximum number of items for a bundle is {cls.MAXIMUM_ITEMS_PER_BUNDLE}")

        if (
            len(tickers) * len(symbols) <= cls.MAXIMUM_ITEMS_PER_REQUEST
            and len(tickers) <= cls.MAXIMUM_INSTRUMENTS_PER_REQUEST
        ):
            yield tickers, symbols
        else:
            nb_tickers = min(
                math.floor(cls.MAXIMUM_ITEMS_PER_REQUEST / len(symbols)),
                len(tickers),
                cls.MAXIMUM_INSTRUMENTS_PER_REQUEST,
            )
            yield tickers[0:nb_tickers], symbols
            yield from cls._breakdown_requests(tickers[nb_tickers:], symbols)

    def get_last_fx_rate(self, base_currency_key: str, target_currency_key: str):
        last_currency_rate = CurrencyFXRates.objects.latest("date").date
        try:
            base_currency = Currency.objects.get(key=base_currency_key)
            target_currency = Currency.objects.get(key=target_currency_key)
            return float(base_currency.convert(last_currency_rate, target_currency, exact_lookup=True))
        except Currency.DoesNotExist:
            return 1.0

    def _process_raw_requests(self, tickers: list[str], fields: list[str], **extra_client_kwargs) -> pd.DataFrame:
        """
        Utility function to fetch data in bulk, aggregate and clean the result
        Args:
            tickers: Instruments tickers
            fields: Symbols
            **extra_client_kwargs: Extra keyword arguments to be passed down to the client (e.g. frequency)

        Returns:
            An aggregated and cleaned dataframe with [Dates, Instrument] as index for timeserie and [Instrument] As index for static data
        """
        requests_data = list(self._breakdown_requests(tickers, fields))
        reqs = []
        if len(requests_data) > self.MAXIMUM_REQUESTS_PER_BUNDLE:
            raise ValueError(f"number of request exceed {self.MAXIMUM_REQUESTS_PER_BUNDLE}")
        # Construct the requests bundle
        for request_tickers, _ in requests_data:
            # Convert a list of string into a valid string
            converted_ticker = ",".join(request_tickers)
            if "start" in extra_client_kwargs or "end" in extra_client_kwargs:
                reqs.append(
                    self.connection.post_user_request(tickers=converted_ticker, fields=fields, **extra_client_kwargs)
                )
            else:
                reqs.append(self.connection.post_user_request(tickers=converted_ticker, fields=fields, kind=0))
        # concat the bundle results
        res = self.connection.get_bundle_data(bundleRequest=reqs)

        df = pd.DataFrame()
        if res:
            if "start" in extra_client_kwargs or "end" in extra_client_kwargs:
                res = list(filter(lambda subdf: subdf.index.name == "Dates", res))
            if res:
                df = pd.concat(res)
                if df.index.name == "Dates":
                    df = (
                        pd.melt(df, ignore_index=False)
                        .reset_index()
                        .rename(columns={"value": "Value", "Field": "Datatype"})
                    )
                df.Value = df.Value.apply(lambda x: None if str(x).startswith("$$ER") else x)
                if not df.empty:
                    df = df[df.Value.notnull()]
                    df = df.replace({"NA": None})
                    if "Dates" in df.columns:
                        df.Dates = pd.to_datetime(df.Dates, utc=True)
                        indexes = ["Instrument", "Dates"]
                    else:
                        indexes = ["Instrument"]
                    df = pd.pivot_table(
                        df, values="Value", index=indexes, columns="Datatype", aggfunc="first", dropna=False
                    )
        return df

    def _normalize_df_units(
        self,
        df: pd.DataFrame,
        ibes_non_per_share_fields: list[str] | None = None,
        ibes_currency_based_fields: list[str] | None = None,
        ibes_fields: list[str] | None = None,
        **extra_client_kwargs,
    ) -> pd.DataFrame:
        """
        Datastream use arbitrary units for some symbols where the denominator needs to be fetch in two steps:
        * Using FIELD#U to get the datastream unit
        * if ibes_non_per_share_fields is provided, fetch for the static symbol IBUNIT and normalized the given columns with the result times a constant (100000)
        * if ibes_currency_based_fields is provided, fetch for the static symbol IBCUR and normalized the given columns latest found currency rate.
        This function uses the index named "Instrument" as ticker fields and the columns as symbols. Start, End or Frequency needs to be passed down using the extra_client_kwargs argument

        Args:
            df: The dataframe to be normalized
            ibes_non_per_share_fields: Columns that need ibunit normalization. Defaults to None
            ibes_currency_based_fields: Columns that need IBES Currency conversion from static field IBCUR
            **extra_client_kwargs: Extra keywords arguments to be passed down to the client (e.g. frequency)

        Returns:
            A normalized dataframe (e.g. volume are in shares and not a million of shares)
        """
        if not ibes_fields:
            ibes_fields = []
        df = df.copy()
        fields = df.columns.unique().tolist()
        tickers = df.index.get_level_values("Instrument").unique().tolist()
        dfu = self._process_raw_requests(tickers, list(map(lambda x: x + "#U", fields)), **extra_client_kwargs).fillna(
            1
        )
        dfu = dfu[dfu.index.isin(df.index)]
        if not dfu.empty:
            dfu = dfu.rename(columns=lambda x: x.replace("#U", "")).fillna(1)
            dfu = dfu[dfu.columns.intersection(df.columns)]
            dff = df[dfu.columns].multiply(dfu, fill_value=1)

            # We do this to ensure that not provided data (nan) are not set with the multiplication with the U matrix
            dff[df.isnull()] = df[df.isnull()]
            dff[df.columns.difference(dff.columns)] = df[
                df.columns.difference(dff.columns)
            ]  # We ensure that the inital colums from df are appended to the new dataframe if missing from dfu. (happens when the colum has non number values)
            df = dff

            # If the symbol is computed directly from refinitiv, we need to normalize it as well given the formula (define as the field name)
            for mav_field in list(filter(lambda x: "MAV#" in x, fields)):
                re_matches = re.findall(r"X\(([^\)]+)\)", mav_field)
                if len(re_matches) > 0 and re_matches[0] in dfu.columns:
                    df.loc[:, mav_field] = df.loc[:, mav_field] * dfu.loc[:, re_matches[0]]

        if ibes_non_per_share_fields:
            df_ibunit = self.get_static_df(tickers, ["IBUNIT"]).fillna(1)
            if not df_ibunit.empty:
                df_ibunit = df_ibunit.set_index("Instrument")["IBUNIT"]
                df[df.columns.intersection(ibes_non_per_share_fields)] = (
                    df[df.columns.intersection(ibes_non_per_share_fields)].multiply(
                        df_ibunit, axis=0, level="Instrument"
                    )
                    * self.IBUNIT_DEFAULT_UNIT
                )
                ibes_fields = list(set(ibes_fields + ibes_non_per_share_fields))
        if ibes_currency_based_fields:
            df_ibcur = self.get_static_df(tickers, ["IBCUR", "ISOCUR"]).replace(
                "BPN", "GBX"
            )  # BPN == GBX but our db only support the latter
            if not df_ibcur.empty:
                df_ibcur["rate"] = 1
                different_curr_idx = df_ibcur["IBCUR"] != df_ibcur["ISOCUR"]
                df_ibcur.loc[different_curr_idx, "rate"] = df_ibcur.loc[different_curr_idx].apply(
                    lambda x: self.get_last_fx_rate(x["IBCUR"], x["ISOCUR"]), axis=1
                )
                df[df.columns.intersection(ibes_currency_based_fields)] = df[
                    df.columns.intersection(ibes_currency_based_fields)
                ].multiply(df_ibcur.set_index("Instrument")["rate"], axis=0, level="Instrument")
                ibes_fields = list(set(ibes_fields + ibes_currency_based_fields))
        if ibes_fields:
            df[df.columns.intersection(ibes_fields)] = (
                df[df.columns.intersection(ibes_fields)] * self.IBUNIT_DEFAULT_UNIT
            )
        return df

    def get_static_df(self, tickers: List[str], fields: List[str], **kwargs) -> pd.DataFrame:
        """
        Public function to returns a dataframe for static symbols

        Args:
            tickers: Ticker to fetch
            fields: Static symbols to fetch

        Returns:
            A valid dataframe result
        """
        # Breakdown tickers and fields into a valid datastream parameters subsets
        return self._process_raw_requests(tickers, fields).reset_index()

    def get_timeserie_df(
        self,
        tickers: List[str],
        fields: List[str],
        ibes_non_per_share_fields: Optional[list[str]] = None,
        ibes_currency_based_fields: Optional[list[str]] = None,
        ibes_fields: list[str] | None = None,
        **extra_client_kwargs,
    ) -> pd.DataFrame:
        """
        Public function to get timeserie type data
        Args:
            tickers: Instrument tickers
            fields: Symbols to fetch
            ibes_non_per_share_fields: Columns that need ibunit normalization. Defaults to None
            ibes_currency_based_fields: Columns that need IBES Currency conversion from static field IBCUR
            **extra_client_kwargs: Extra keywords arguments to be passed down to the client (e.g. frequency)

        Returns:
            The result as a dataframe
        """
        if (start := extra_client_kwargs.get("start", None)) and isinstance(start, date):
            extra_client_kwargs["start"] = start.strftime("%Y-%m-%d")
        if (end := extra_client_kwargs.get("end", None)) and isinstance(end, date):
            extra_client_kwargs["end"] = end.strftime("%Y-%m-%d")

        final_df_list = []
        # Sometime, we get too many fields and therefore, we need to split them based on the maximum number of symbols allowed per request
        for splited_fields in np.array_split(fields, math.ceil((len(fields) / self.MAXIMUM_DATATYPES_PER_REQUEST))):
            # Breakdown tickers and fields into a valid datastream parameters subsets
            df = self._process_raw_requests(tickers, list(splited_fields), **extra_client_kwargs)
            if not df.empty:
                df = self._normalize_df_units(
                    df,
                    ibes_non_per_share_fields=ibes_non_per_share_fields,
                    ibes_currency_based_fields=ibes_currency_based_fields,
                    ibes_fields=ibes_fields,
                    **extra_client_kwargs,
                )
                df = (
                    df.reset_index()
                    .rename_axis(None)
                    .replace([np.inf, -np.inf, np.nan], None)
                    .dropna(how="all", subset=df.columns.intersection(splited_fields))
                )
                if not df.empty:
                    if "Dates" in df.columns:
                        df = df.set_index(["Instrument", "Dates"])
                    else:  # otherwise it's not a timeseries and we set index only on instrument
                        df = df.set_index(["Instrument"])
                    final_df_list.append(df)
        if final_df_list:
            return pd.concat(final_df_list, axis=1).reset_index()
        return pd.DataFrame()

    def raw_fetch(
        self, tickers: List[str], fields: List[str], start: Optional[date] = None, end: Optional[date] = None
    ):
        """
        Utility function to expose the client directly without any modification
        """
        return self.connection.get_data(tickers=tickers, fields=fields, start=start, end=end)



# Generated by Django 5.2.9 on 2025-12-16 15:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbfdm', '0035_alter_instrument_related_instruments_and_more'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='relatedinstrumentthroughmodel',
            unique_together=set(),
        ),

    ]
